from ast import literal_eval
from numpy import float64, array, append
class Set:
	"""
	Class representing a set (training set / test set).
	"""
	def __init__(self, sequences:list, times:list = None, t: int = None) -> None:
		"""
		Creates the Set.

		Parameters
		----------
		sequences : list
			List of sequences of traces.
			A trace can be:
			- a sequence of observations (list of str) ->HMM or MC
			- an alternating sequence of actions/observations (list of str) ->MDP
			- a sequence of timed observations (list of float) ->GOHMM
			- a sequence of vectors of timed observations (list of list of float) ->MGOHMM
			- an alternating sequence of waiting-times/observations (list of float and str) ->CTMC
		times : list, optional
			`times[i]` is the number of time `sequences[i]` appears in the set.
			Can be omitted.
		t : int, optional
			- 0 if this set was generated by a HMM or a MC,
			- 1 if it was generated by a MDP,
			- 3 if it was generated by a GOHMM,
			- 4 if it was generated by a CTMC.
		"""
		if sequences == []:
			self.sequences = []
			self.times = []
			self.type = t
			return
		if t != None:
			self.type = t
		else:
			if type(sequences[0][0])  == float64 or type(sequences[0][0])  == float:
				self.type = 2 # GOHMM
			elif (type(sequences[0][1])  == float64 or type(sequences[0][1])  == float) and type(sequences[0][0])  == str:
				self.type = 4 # timed CTMC
			elif type(sequences[0][0])  == list:
				self.type = 3 # MGOHMM
			else:
				self.type = 0 # HMM or MC or non-timed CTMC

		if type(times) == type(None):
			self.setFromList(sequences)
		else:
			self.sequences = sequences
			self.times = array(times)
	
	def save(self,file_path: str) -> None:
		"""
		Save this set into a text file.
		
		Parameters
		----------
		file_path: str
			where to save
		"""
		f = open(file_path,'w')
		f.write(str(self.type)+'\n')
		for i in range(len(self.sequences)):
			f.write(str(self.sequences[i])+'\n')
			f.write(str(self.times[i])+'\n')
		f.close()

	def setFromList(self, l: list) -> None:
		"""
		Convert a list of sequences of observations to a set.

		Parameters
		----------
		l : list
			list of sequences of observations.
		"""
		res = [[],[]]
		for s in l:
			s = list(s)
			if s not in res[0]:
				res[0].append(s)
				res[1].append(0)
			res[1][res[0].index(s)] += 1
		self.sequences = res[0]
		self.times = array(res[1])
	
	def isEqual(self, set2) -> bool:
		"""
		Checks wheter or not this set is equal to another set `set2`.

		Parameters
		----------
		set2 : Set
			Another set.

		Returns
		-------
		bool
			True if this set is equal to `set2`.
		"""
		if self.type != set2.type:
			return False
		if sum(self.times) != sum(set2.times):
			return False
		if len(self.sequences) != len(set2.sequences):
			return False
		if len([i for i in self.sequences if i in set2.sequences]) != len(set2.sequences):
			return False
		for s,t1 in zip(self.sequences,self.times):
			try:
				t2 = set2.times[set2.sequences.index(s)]
			except ValueError:
				return False
			if t1 != t2:
				return False
		return True
	
	def getAlphabet(self) -> list:
		"""
		Returns the list of all possible observations in this set.

		Returns
		-------
		list
			list of observations.
		"""
		if self.type == 2 or self.type == 3:
			return [] #doesn't make sense
		observations = []
		if self.type == 4: # timed self.sequences
			for sequence_obs in self.sequences:
				for x in range(0,len(sequence_obs),2):
					if sequence_obs[x] not in observations:
						observations.append(sequence_obs[x])	
		elif self.type == 0:
			for sequence_obs in self.sequences: # non-timed self.sequences
				for x in sequence_obs:
					if x not in observations:
						observations.append(x)
		elif self.type == 1:
			return self.getActionsObservations()[1]

		return observations
	

	def getActionsObservations(self) -> list:
		"""
		Returns all possible observations and all possible actions in this set.

		Returns
		-------
		list
			list of one list of actions and one list of observations.
		"""
		if self.type != 1:
			return self.getAlphabet()
		actions = []
		observations = []
		for seq in range(len(self.sequences)):
			actions += [self.sequences[seq][i+1] for i in range(0,len(self.sequences[seq])-1,2)]
			observations += [self.sequences[seq][i] for i in range(0,len(self.sequences[seq])-1,2)] + [self.sequences[seq][-1]]
		actions = list(set(actions))
		observations = list(set(observations))
		return [actions,observations]
	
	def addSet(self, s2):
		"""
		Merges this set with another generated by the same kind of Markov model.

		Parameters
		----------
		s2 : Set
			set 2.
		
		Raises
		------
		TypeError
			If the two models `type` attribute are different.
		"""
		if self.type != s2.type:
			raise TypeError("The two sets should have been generated by the same"+
							"kind of Markov model.")
		for i in range(len(s2.sequences)):
			if not s2.sequences[i] in self.sequences:
				self.sequences.append(s2.sequences[i])
				self.times = append(self.times, s2.times[i])
			else:
				self.times[self.sequences.index(s2.sequences[i])] += s2.times[i]

def loadSet(file_path: str) -> Set:
	"""
	Load a training/test set saved into a text file.

	Parameters
	----------
	file_path: str
		location of the file.
	
	Returns
	-------
	Set
		a training/test set.
	"""
	res_set = [[],[]]
	f = open(file_path,'r')
	l = f.readline()
	t = literal_eval(l[:-1])
	l = f.readline()
	while l:
		res_set[0].append(literal_eval(l[:-1]))
		l = f.readline()
		res_set[1].append(int(l[:-1]))
		l = f.readline()
	f.close()
	return Set(res_set[0],res_set[1],t)



from math import  log
from .tools import resolveRandom
from .Set import Set
from numpy import array, where, append, dot, zeros, vsplit
from numpy.random import geometric
from multiprocessing import cpu_count, Pool
from sys import platform
from math import log

MC_ID   = 0
MDP_ID  = 1
HMM_ID  = 2
GOHMM_ID= 3
CTMC_ID = 4
PCTMC_ID= 5

class Model:
	"""
	Abstract class that represents a model.
	Is inherited by all the classes representing any model.
	Should not be instantiated itself!
	"""
	def __init__(self,matrix,initial_state,name: str) -> None:
		"""
		Creates an abstract model.

		Parameters
		----------
		matrix: ndarray or dict
			The transition matrix.

		initial_state: int or list of float
			determine which state is the initial one (then it's the id of the
			state), or what are the probability to start in each state
			(then it's a list of probabilities).

		name: str
			name of the model.
		"""
		self.nb_states = len(matrix)
		# initial_state can be a list of probability or an int
		if type(initial_state) == int:
			self.initial_state = array([0.0 for i in range(self.nb_states)])
			self.initial_state[initial_state] = 1.0
		else:
			if round(sum(initial_state)) != 1.0:
				msg = "Error: the sum of initial_state should be 1.0, here it's"
				msg+= str(sum(initial_state))
				raise ValueError(msg)
			self.initial_state = array(initial_state)
		self.matrix = matrix
		self.name = name
	
	def rename(self,name: str) -> None:
		"""
		Change the name of the model.

		Parameters
		----------
		name : str
			new name.
		"""
		self.name = name

	def next(self,state: int):
		"""
		overrided
		"""
		pass

	def _stateToString(self,state:int):
		"""
		overrided
		"""
		pass


	def _save(self,f) -> None:
		f.write(self.name)
		f.write('\n')
		f.write(str(self.initial_state.tolist()))
		f.write('\n')
		f.write(str(self.matrix.tolist()))
		f.write('\n')
		f.close()

	def tau(self,s1: int,s2: int,obs) -> float:
		"""
		overrided
		"""
		pass

	def pi(self,s: int) -> float:
		"""
		Return the probability of starting in state ``s``.

		Parameters
		----------
		s: int
			state ID.

		Returns
		-------
		output : float
			the probability of starting in state s.
		"""
		return self.initial_state[s]

	def run(self, number_steps: int, current: int = -1) -> list:
		"""
		Simulates a run of length ``number_steps`` of the model and return the
		sequence of observations generated.
		
		Parameters
		----------
		number_steps: int
			length of the simulation.
		current : int, optional.
			If current it set, it starts from the state `current`.
			Otherwise it starts from an initial state.

		Returns
		-------
		output: list of str
			trace generated by the run.
		"""
		output = []
		if current == -1:
			current = resolveRandom(self.initial_state)

		while len(output) < number_steps:
			[next_state, symbol] = self.next(current)
			output.append(symbol)
			current = next_state
		
		return output
	
	def generateSet(self, set_size: int, param, distribution=None, min_size=None, timed: bool=False) -> Set:
		"""
		Generates a set (training set / test set) containing ``set_size`` traces.

		Parameters
		----------
		set_size: int
			number of traces in the output set.
		param: a list, an int or a float.
			the parameter(s) for the distribution. See "distribution".
		distribution: str, optional
			If ``distribution=='geo'`` then the sequence length will be
			distributed by a geometric law such that the expected length is
			``min_size+(1/param)``.
			If distribution==None param can be an int, in this case all the
			seq will have the same length (``param``), or ``param`` can be a
			list of int.
			Default is None.
		min_size: int, optional
			see "distribution". Default is None.
		timed: bool, optional
			Only for timed model. Generate timed or non-timed traces.
			Default is False.
		
		Returns
		-------
		output: Set
			a set (training set / test set).
		
		Examples
		--------
		>>> set1 = model.generateSet(100,10)
		>>> # set1 contains 100 traces of length 10
		>>> set2 = model.generate(100, 1/4, "geo", min_size=6)
		>>> # set2 contains 100 traces. The length of the traces is distributed following
		>>> # a geometric distribution with parameter 1/4. All the traces contains at
		>>> # least 6 observations, hence the average length of a trace is 6+(1/4)**(-1) = 10.
		"""
		seq = []
		val = []
		for i in range(set_size):
			if distribution == 'geo':
				curr_size = min_size + int(geometric(param))
			else:
				if type(param) == list:
					curr_size = param[i]
				elif type(param) == int:
					curr_size = param

			if timed:
				trace = self.run(curr_size,timed)
			else:
				trace = self.run(curr_size)

			if not trace in seq:
				seq.append(trace)
				val.append(1)
			else:
				val[seq.index(trace)] += 1

		return Set(seq,val)


	def __str__(self) -> str:
		res = "Name: "+self.name+'\n'
		if 1.0 in self.initial_state:
			res += "Initial state: s"+str(where(self.initial_state==1.0)[0][0])+'\n'
		else:
			res += "Initial state: "
			for i in range(len(self.matrix)):
				if self.initial_state[i] >= 0.001:
					res += 's'+str(i)+': '+str(round(self.initial_state[i],3))+', '
			res = res[:-2]+'\n'
		for i in range(self.nb_states):
			res += self._stateToString(i)+'\n'
		res += '\n'
		return res

	def logLikelihood(self,sequences: Set) -> float:
		"""
		Compute the average loglikelihood of a set.

		Parameters
		----------
		sequences: Set
			A set.
		
		Returns
		-------
		output: float
			loglikelihood of ``sequences`` under this model.
		
		Examples
		--------
		>>> model.logLikelihood(set1)
		-4.442498878506513
		"""
		if platform != "win32" and platform != "darwin":
			return self._logLikelihood_multiproc(sequences)
		else:
			return self._logLikelihood_oneproc(sequences)

	def _updateAlphaMatrix(self,sequence: list,
						   common: int,
						   alpha_matrix: list) -> array:
		"""
		Update the given alpha values for all the states for a new ``sequence``
		of observations. It keeps the alpha values for the ``common`` first
		observations of the ``sequence``. The idea is the following: if you
		have already computed the alpha values for a previous sequence and you
		want to compute the alpha values of a new sequence that starts with the
		same ``common`` observations you don't need to compute again the first
		``common`` alpha values for each states. If, on the other hand, you
		have still not computed any alpha values you can simply set ``common``
		to 0 and give an empty ``alpha_matrix`` which has the right size.
		The method ``initAlphaMatrix`` can generate such matrix.

		Parameters
		----------
		sequence: list of str
			the sequence of observations for which we compute the alpha values.
		common: int
			for each state, the first ``common`` alpha values will be keept
			unchanged.
		alpha_matrix: 2-D narray of float
			the ``alpha_matrix`` to update. Can be generated by the method
			``initAlphaMatrix``.

		Returns
		-------
		output: 2-D narray of float
			the alpha matrix containing all the alpha values for all the states
			for this sequence: ``alpha_matrix[s][t]`` is the probability of
			being in state ``s`` after seing the ``t-1`` first observation of
			``sequence``.
		"""
		diff_size = len(alpha_matrix)-1 - len(sequence)
		if diff_size < 0: # alpha_matrix too small
			n = zeros(-diff_size * self.nb_states).reshape(-diff_size,self.nb_states)
			alpha_matrix = append(alpha_matrix,n,axis=0)
		elif diff_size > 0: #alpha_matrix too big
			alpha_matrix = vsplit(alpha_matrix,[len(alpha_matrix)-diff_size,self.nb_states])[0]
		for k in range(common,len(sequence)):
			for s in range(self.nb_states):
				p = array([self.tau(ss,s,sequence[k]) for ss in range(len(self.matrix))])
				alpha_matrix[k+1,s] = dot(alpha_matrix[k],p)
		return alpha_matrix

	def _initAlphaMatrix(self,len_seq: int) -> array:
		"""
		Return a matrix with the correct size to be updated by the
		``updateAlphaMatrix`` method.

		Parameters
		----------
		len_seq: int
			length of the sequence for which we will compute the alpha values.

		Returns
		-------
		output: a matrix with the correct size to be updated by the updateAlphaMatrix method
		:rtype: 2-D narray of float
		"""
		zero_arr = zeros(shape=(len_seq*self.nb_states,))
		alpha_matrix = append(self.initial_state,zero_arr).reshape(len_seq+1,self.nb_states)
		return alpha_matrix

	def _logLikelihood_oneproc(self,sequences: Set) -> float:
		"""
		Compute the average loglikelihood of a set of sequences if
		multiprocessing is disable.

		Parameters
		----------
		sequences: Set
			A set.
		
		Returns
		-------
		output: float
			loglikelihood of ``sequences`` under this model.
		"""
		sequences_sorted = sequences.sequences[:]
		sequences_sorted.sort()
		loglikelihood = 0.0
		alpha_matrix = self._initAlphaMatrix(len(sequences_sorted[0]))
		for seq in range(len(sequences_sorted)):
			sequence = sequences_sorted[seq]
			times = sequences.times[sequences.sequences.index(sequence)]
			common = 0
			if seq > 0:
				while common < min(len(sequences_sorted[seq-1]),len(sequence)):
					if sequences_sorted[seq-1][common] != sequence[common]:
						break
					common += 1
					
			alpha_matrix = self._updateAlphaMatrix(sequence,common,alpha_matrix)
			if alpha_matrix[-1].sum() > 0:
				loglikelihood += log(alpha_matrix[-1].sum()) * times

		return loglikelihood / sum(sequences.times)

	def _logLikelihood_multiproc(self, sequences: Set) -> float:
		p = Pool(processes = cpu_count()-1)
		tasks = []
		for seq,times in zip(sequences.sequences,sequences.times):
			tasks.append(p.apply_async(self._computeAlphas, [seq, times,]))
		temp = [res.get() for res in tasks if res.get() != False]
		s = sum(temp)
		if s == 0.0:
			print('WARNING: the model is not able to generate any sequence in the set')
			return 0.0
		return s/sum(sequences.times)

	def _computeAlphas(self,sequence: list, times: int) -> float:
		"""
		Compute the alpha values for ``sequence``.

		Parameters
		----------
		sequence: list of str
			Sequence of observations.
		times: int
			Number of times this sequence appears in the sample.

		Returns
		-------
		float
			loglikelihood of ``sequence`` multiplied by ``times``.
		"""
		len_seq = len(sequence)
		prev_arr = self.initial_state
		for k in range(len_seq-1):
			new_arr = zeros(self.nb_states)
			for s in range(self.nb_states):
				p = array([self.tau(ss,s,sequence[k]) for ss in range(self.nb_states)])
				new_arr[s] = dot(prev_arr,p)
			prev_arr = new_arr
		prev_arr *= (array([self.b(s,sequence[-1]) for s in range(self.nb_states)]))
		try:
			res = log(prev_arr.sum())*times
		except ValueError:
			print("WARNING: this model is not able to generate",sequence)
			res = False
		return res
	
	def _checkStateIndex(self,s:int) -> None:
		if type(s) != int:
			raise TypeError('The parameter must be a valid state ID')
		elif s < 0:
			raise IndexError('The parameter must be a valid state ID')
		elif s >= self.nb_states:
			raise IndexError('This model contains only '+str(self.nb_states)+' states')


import unittest
from ..ctmc import *
from ..pctmc import loadPCTMC
from os import remove
from ..base.Set import *
from math import exp
from numpy import where, array
from ..base.BW import BW

def modelCTMC1():
	labelling = ['red','red','yellow','blue','blue']
	transitions = [(0,1,0.08),(0,2,0.12),(1,1,0.3),(1,2,0.7),
				   (2,0,0.2),(2,3,0.1),(2,4,0.2),(3,3,0.8),
				   (3,1,0.1),(3,4,0.1),(4,2,0.25)]

	return createCTMC(transitions,labelling,0,"My_CTMC")

def modelCTMC_car_tl():
	labelling = ['red','green','orange']
	transitions = [(2,0,2.0)]
	sync_trans = [(0,'ped_red',1,0.5),
				  (1,'button',2,1.0)]
	return createCTMC(transitions,labelling,[0.6,0.4,0.0],"Car_tl",sync_trans)

def modelCTMC_ped_tl():
	labelling = ['red','green']
	transitions = [(1,0,0.1)]
	sync_trans = [(0,'button',1,0.5),
				  (0,'ped_red',0,10.0)]
	return createCTMC(transitions,labelling,0,"Pedestrian_tl",sync_trans)


m1 = modelCTMC1()

class CTMCTestclass(unittest.TestCase):
	
	def test_CTMC_initial_state(var):
		labelling = ['red','red','yellow','blue','blue']
		transitions = [(0,1,0.08),(0,2,0.12),(1,1,0.3),(1,2,0.7),
				   (2,0,0.2),(2,3,0.1),(2,4,0.2),(3,3,0.8),
				   (3,1,0.1),(3,4,0.1),(4,2,0.25)]
		ctmc = createCTMC(transitions,labelling,0)
		var.assertEqual(ctmc.nb_states,6)
		var.assertEqual(ctmc.labelling.count('init'),1)
		var.assertEqual(ctmc.getLabel(int(where(ctmc.initial_state == 1.0)[0][0])),'init')
		
		labelling=['a','b','c','d','a']
		ctmc = createCTMC(transitions,labelling,[0.3,0.0,0.0,0.2,0.5])
		var.assertEqual(ctmc.nb_states,6)
		var.assertEqual(ctmc.labelling.count('init'),1)
		var.assertEqual(ctmc.pi(5),1.0)
		var.assertTrue((ctmc.matrix[-1]==array([0.3,0.0,0.0,0.2,0.5,0.0])).all())
		
		labelling=['a','b','c','d','a']
		ctmc = createCTMC(transitions,labelling,array([0.3,0.0,0.0,0.2,0.5]))
		var.assertEqual(ctmc.nb_states,6)
		var.assertEqual(ctmc.labelling.count('init'),1)
		var.assertEqual(ctmc.pi(5),1.0)
		var.assertTrue((ctmc.matrix[-1]==array([0.3,0.0,0.0,0.2,0.5,0.0])).all())
		
	def test_CTMC_state(var):
		var.assertEqual(m1.e(0),1/5)
		var.assertEqual(m1.expected_time(0),5.0)
		var.assertEqual(m1.l(0,2,'red'),0.12)
		var.assertEqual(m1.l(0,3,'red'),0.0)
		var.assertEqual(m1.tau(0,2,'red'),0.6)
		var.assertEqual(m1.tau(0,2,'yellow'),0.0)
		t = 1.0
		lkl = 0.2*exp(-0.2*t)
		var.assertEqual(m1.lkl(0,t),lkl)
		var.assertEqual(m1.lkl(0,-t),0.0)
		var.assertEqual(set(m1.getAlphabet()),
						set(['red','yellow','blue','init']))
		var.assertEqual(m1.getLabel(2),'yellow')
	
	def test_CTMC_save_load_str(var):
		m1.save("test_save.txt")
		mprime = loadCTMC("test_save.txt")
		var.assertEqual(str(m1),str(mprime))
		remove("test_save.txt")
	
	def test_CTMC_Set(var):
		set1 = m1.generateSet(50,10,timed=True)
		var.assertEqual(set1.type,4)
		set2 = m1.generateSet(50,1/4,"geo",6, timed=True)
		set1.addSet(set2)
		var.assertEqual(set1.type,4)
		set1.save("test_save.txt")
		set2 = loadSet("test_save.txt")
		var.assertEqual(set2.type,4)
		var.assertTrue(set1.isEqual(set2))
		remove("test_save.txt")
	
	def test_CTMC_logLikelihood(var):
		set1 = Set([['init','red','red','red','yellow','blue']],[1])
		l11 = m1._logLikelihood_oneproc(set1)
		l12 = m1._logLikelihood_multiproc(set1)
		var.assertAlmostEqual(l11,l12)
		var.assertAlmostEqual(l11,-2.9877641039048144)

		set2 = Set([['init',0.5,'red', 0.9882294310276254, 'red', 0.030426756030141687,
					'yellow', 7.774379967385992, 'red', 13.783896165945656,
					'yellow', 0.09826006650405217, 'red'],
					['init',0.5,'red', 3.5876459831883487, 'yellow', 0.21449179923772158,
					'red', 1.2177338376867959, 'red', 0.4591551146146921,
					'yellow', 1.25461684148389, 'red']],[1,1])
		l21 = m1.logLikelihood(set2)
		var.assertAlmostEqual(l21,-13.259521595075453)
	
	def test_BW_CTMC(var):
		initial_model   = loadCTMC("jajapy/tests/materials/ctmc/random_CTMC.txt")
		
		training_set    = loadSet("jajapy/tests/materials/ctmc/training_set_timed_CTMC.txt")
		output_expected = loadCTMC("jajapy/tests/materials/ctmc/output_timed_CTMC.txt")
		output_gotten   = BW().fit( training_set, initial_model, stormpy_output=False)
		test_set = m1.generateSet(10000,10)
		var.assertAlmostEqual(output_expected.logLikelihood(test_set),
							  output_gotten.logLikelihood(test_set))
		
		training_set    = loadSet("jajapy/tests/materials/ctmc/training_set_nontimed_CTMC.txt")
		output_expected = loadCTMC("jajapy/tests/materials/ctmc/output_nontimed_CTMC.txt")
		output_gotten   = BW().fit( training_set, initial_model, stormpy_output=False)
		test_set = m1.generateSet(10000,10)
		var.assertAlmostEqual(output_expected.logLikelihood(test_set),
							  output_gotten.logLikelihood(test_set))
	
	def test_sync_composition(var):
		output_gotten = synchronousCompositionCTMCs(modelCTMC_car_tl(),modelCTMC_ped_tl())
		output_expected = loadPCTMC("jajapy/tests/materials/ctmc/composition.txt")
		var.assertEqual(str(output_expected),str(output_gotten))

if __name__ == "__main__":
	unittest.main()


from sys import platform
from multiprocessing import cpu_count, Pool
from numpy.polynomial.polynomial import polyroots
from numpy import array, dot, append, zeros, ones, float64, inf, ndarray, log, isnan, sqrt, stack, longdouble, float16, where, delete, newaxis, full
from math import exp
from datetime import datetime
from .Set import Set
from alive_progress import alive_bar
from .Model import MC_ID, MDP_ID, CTMC_ID, PCTMC_ID, HMM_ID, GOHMM_ID
from sympy import sympify

NB_PROCESS = cpu_count()-1

class BW:
	"""
	Class for the Baum-Welch algorithm.
	"""
	def __init__(self):
		try:
			from ..with_stormpy import jajapyModeltoStormpy, stormpyModeltoJajapy
			self.jajapyModeltoStormpy = jajapyModeltoStormpy
			self.stormpyModeltoJajapy = stormpyModeltoJajapy
			self.stormpy_installed = True
		except ModuleNotFoundError:
			self.stormpy_installed = False


	def fit(self,training_set,initial_model=None, nb_states: int=None,
			random_initial_state: bool=True, min_exit_rate_time : int=None,
			max_exit_rate_time: int=None, self_loop: bool =None,
			nb_distributions: int=None, output_file: str=None,
			output_file_prism: str=None, epsilon: float=0.01, max_it: int=inf,
			pp: str='', verbose: bool=True, return_data: bool=False,
			stormpy_output: bool = True,fixed_parameters: ndarray = False,
			update_constant :bool = True, min_val: float = None, max_val: float = None):
		"""
		Fits any model according to ``traces``.
		This method will figure which type of Markov model should be used, according to the
		training set and the initial model (if given).

		Parameters
		----------
		traces : Set, list or ndarray.
			The training set.
		initial_model : Model or stormpy.sparse model, optional.
			The first hypothesis. If not set it will create a random Model
			with ``nb_states`` states.
		nb_states: int, optional.
			If ``initial_model`` is not set it will create a random Model with
			``nb_states`` states. Must be set if ``initial_model`` is not set.
		random_initial_state: bool, optional.
			If ``initial_model`` is not set it will create a random Model with
			random initial state according to this sequence of probabilities.
			Must be set if ``initial_model`` is not set.
			Default is True.
		min_exit_rate_time: int, optional
			For CTMC learning only.
			Minimum exit rate for the states in the first hypothesis.
			Must be set if ``initial_model`` is not set, and if the SUL is a
			CTMC or PCTMC.
		max_exit_rate_time: int, optional
			For CTMC learning only.
			Minimum exit rate for the states in the first hypothesis.
			Must be set if ``initial_model`` is not set, and if the SUL is a
			CTMC or PCTMC.
		self_loop: bool, optional
			For CTMC/PCTMC learning only.
			Wether or not there will be self loop in the first hypothesis.
			Must be set if ``initial_model`` is not set, and if the SUL is a
			CTMC or PCTMC.
		nb_distributions: int, optional.
			For GoHMM learning only.
			Number of distributions in each state in the initial hypothsis.
			Must be set if ``initial_model`` is not set, and if the SUL is a
			GoHMM.
		output_file : str, optional
			If set, the output model will be saved at this location.
			Otherwise the output model will not be saved.
		output_file_prism : str, optional
			If set, the output model will be saved in a prism file at this
			location. Otherwise the output model will not be saved.
			This parameter is ignored if the model under learning is a HMM
			or a GoHMM.
		epsilon : float, optional
			The learning process stops when the difference between the
			loglikelihood of the training set under the two last hypothesis is
			lower than ``epsilon``. The lower this value the better the output,
			but the longer the running time.
			Default is 0.01.
		max_it: int, optional
			Maximal number of iterations. The algorithm will stop after `max_it`
			iterations.
			Default is infinity.
		pp:	str, optional
			Will be printed at each iteration.
			Default is an empty string.
		verbose: bool, optional.
			Print or not a small recap at the end of the learning.
			Default is True.
		return_data: bool, optional.
			If set to True, a dictionary containing following values will be
			returned alongside the output model once the learning is done.
			'learning_rounds', 'learning_time', 'training_set_loglikelihood'.
			Default is False.
		stormpy_output: bool, optional.
			If set to True the output model will be a Stormpy sparse model.
			Doesn't work for HMM and GOHMM.
			Default is True.
		fixed_parameters: ndarray of bool, optional
			For CTMC/PCTMC learning only.
			ndarray of bool with the same shape as the transition matrix (i.e
			nb_states x nb_states). If `fixed_parameters[s1,s2] == True`, the
			transition parameter from s1 to s2 will not be changed during the
			learning (it's a fixed parameter).
			By default no parameters will be fixed.
		update_constant: bool, optional
			For PCTMC learning only.
			If set to False, the constant transitions (i.e. tha transition
			that doesn't depend on any parameter) will no be updated.
			Default is True.
		min_val: float, optional
			For PCTMC learning only.
			Minimal value for the randomly instantiated parameters.
			If not set and if the model has at least two instantiated parameters,
			this value is equal to the parameters with the smallest instantiation.
			If not set and if the model has less than two instantiated parameters,
			this value is equal to 0.1.
		max_val : float, optional
			For PCTMC learning only.
			Maximal value for the randomly instantiated parameters.
			If not set and if the model has at least two instantiated parameters,
			this value is equal to the parameters with the highest instantiation.
			If not set and if the model has less than two instantiated parameters,
			this value is equal to 5.0.	

		Returns
		-------
		Model
			fitted model.
		"""

		stormpy_output = self._preparation(training_set,initial_model, nb_states,
						  random_initial_state, min_exit_rate_time,
						  max_exit_rate_time, self_loop,
						  nb_distributions, stormpy_output, 
						  fixed_parameters, update_constant,
						  min_val, max_val)
		return self._bw( max_it,pp,epsilon,output_file,output_file_prism,verbose,
						stormpy_output,return_data)


	def _preparation(self,training_set,initial_model=None, nb_states: int=None,
			random_initial_state: bool=True, min_exit_rate_time : int=None,
			max_exit_rate_time: int=None, self_loop: bool =None,
			nb_distributions: int=None,
			stormpy_output: bool = True,fixed_parameters: ndarray = False,
			update_constant :bool = True, min_val: float = None, max_val: float = None):	

		self._getInitialModel(initial_model,nb_states)
		self._getTrainingSet(training_set)
		if self.h == None:
			self._createInitialModel(nb_states, random_initial_state,
								 min_exit_rate_time, max_exit_rate_time, self_loop,
								 nb_distributions)
		else:
			if ((self.training_set.type != self.type_model) and
			   (self.training_set.type != 0 and self.type_model != 2) and
			   (self.training_set.type != 0 and self.type_model != 4) and
			   (self.training_set.type != 0 and self.type_model != 5) and
			   (self.training_set.type != 4 and self.type_model != 5)):
				msg = "the training set and the initial hypothesis are notcompatible:\n"
				msg+= "the training set is a set of sequences of "
				msg+=['labels','action-label pairs','vector of continuous observations','dwell time-label pairs'][self.training_set.type]
				msg+= "\nand the initial hypothesis is a "+['MC','MDP','HMM','GoHMM','CTMC','PCTMC'][self.type_model]
				raise RuntimeError(msg)
		
		if self.type_model == MC_ID:
			self._update = self._update_MC
			self._processWork  = self._processWork_MC
			print("Learning an MC...")
		
		elif self.type_model == MDP_ID:
			self._h_tau = self._h_tau_MDP
			self._computeAlphas = self._computeAlphas_MDP
			self._computeBetas = self._computeBetas_MDP
			self._update = self._update_MDP
			self._processWork  = self._processWork_MDP
			print("Learning an MDP...")
			self.actions = self.h.actions
		
		elif self.type_model == CTMC_ID:
			self._update = self._update_CTMC
			self._processWork = self._processWork_CTMC
			print("Learning a CTMC...")
			if self.training_set.type == 4:
				self._computeAlphas = self._computeAlphas_timed
				self._computeBetas  = self._computeBetas_timed
			if type(fixed_parameters) == bool:
				self.fixed_parameters = full(self.h.matrix.shape,False)
			else:
				self.fixed_parameters = fixed_parameters
		
		elif self.type_model == PCTMC_ID:
			print("Learning a PCTMC...")
			self._update = self._update_PCTMC
			self._processWork = self._processWork_PCTMC
			self._computeTaus = self._computeTaus_PCTMC
			if min_val != None:
				if max_val != None:
					self.h.randomInstantiation(min_val=min_val, max_val=max_val)
				else:
					self.h.randomInstantiation(min_val=min_val)
			elif max_val != None:
				self.h.randomInstantiation(max_val=max_val)
			else:
				self.h.randomInstantiation()
				
			if self.training_set.type == 4:
				self._computeAlphas = self._computeAlphas_timed
				self._computeBetas  = self._computeBetas_timed
			self.nb_parameters = self.h.nb_parameters
			self.update_constant = update_constant
			self._h_e = self._h_e_PCTMC
			self._h_l = self._h_l_PCTMC
			self._h_tau=self._h_tau_PCTMC
			self._sortParameters(fixed_parameters)

		elif self.type_model == HMM_ID:
			self._update = self._update_HMM
			self._processWork = self._processWork_HMM
			self.alphabet = self.h.getAlphabet()
			print("Learning an HMM...")
		
		elif self.type_model == GOHMM_ID:
			self._update= self._update_GoHMM
			self._processWork = self._processWork_GoHMM
			self.nb_distr = self.h.nb_distributions
			print("Learning a GoHMM...")

		if stormpy_output and not self.stormpy_installed:
			print("WARNING: stormpy not found. The output model will be a Jajapy model")
			stormpy_output = False
		
		return stormpy_output

	def _bw(self,max_it,pp,epsilon,output_file,output_file_prism,verbose,stormpy_output,return_data):
		start_time = datetime.now()
		self.nb_states = self.h.nb_states

		counter = 0
		prevloglikelihood = 10

		if max_it != inf:
			alive_parameter = max_it
		else:
			alive_parameter = 0
		
		with alive_bar(alive_parameter, dual_line=True, title=str(pp)) as bar:
			while counter < max_it:
				self._computeTaus()
				temp = self._runProcesses(self.training_set)
				currentloglikelihood = self._update(temp)
				counter += 1
				bar_text = '   Diff. loglikelihood: '+str(round(currentloglikelihood-prevloglikelihood,5))+' (>'+str(epsilon)+')'
				bar_text+= '   Av. one iteration (s): '+str(round((datetime.now()-start_time).total_seconds()/counter,2))
				bar.text = bar_text
				bar()
				if abs(prevloglikelihood - currentloglikelihood) < epsilon:
					break
				else:
					prevloglikelihood = currentloglikelihood

		running_time = datetime.now()-start_time
		running_time = running_time.total_seconds()

		if output_file:
			self.h.save(output_file)
		
		if output_file_prism and (self.type_model != HMM_ID and self.type_model != GOHMM_ID):
			self.h.savePrism(output_file_prism)
		
		if verbose:
			self._endPrint(counter,running_time)

		if stormpy_output and (self.type_model != HMM_ID and self.type_model != GOHMM_ID):
			self.h = self.jajapyModeltoStormpy(self.h)

		if return_data:
			info = {"learning_rounds":counter,"learning_time":running_time,"training_set_loglikelihood":currentloglikelihood}
			return self.h, info
			
		return self.h
	

	# MC-----------------------------------------------------------------------
	def _processWork_MC(self,sequence,times):
		alpha_matrix = self._computeAlphas(sequence)
		beta_matrix  = self._computeBetas( sequence)
		proba_seq = alpha_matrix.T[-1].sum()
		if proba_seq == 0.0:
			return False
		####################
		den = (alpha_matrix.T[:-1]*beta_matrix.T[:-1]*times/proba_seq).sum(axis=0)			
		num = zeros(shape=(self.nb_states,self.nb_states))
		for s in range(self.nb_states):
			for ss in range(self.nb_states):
				p = array([self._h_tau(s,ss,o) for o in sequence])
				num[s,ss] = dot(alpha_matrix[s][:-1]*p*beta_matrix[ss][1:],times/proba_seq).sum()
		####################
		return [den,num, proba_seq,times]
	
	def _update_MC(self,temp):
		den = array([i[0] for i in temp]).sum(axis=0)
		num = array([i[1] for i in temp]).sum(axis=0)
		lst_proba=array([i[2] for i in temp])
		lst_times=array([i[3] for i in temp])

		currentloglikelihood = dot(log(lst_proba),lst_times)

		for s in range(len(den)):
			if den[s] == 0.0:
				den[s] = 1.0
				num[s] = self.h.matrix[s]

		matrix = num/den[:, newaxis]
		self.h.matrix = matrix
		return currentloglikelihood


	# MDP----------------------------------------------------------------------
	def _h_tau_MDP(self,s1: int,act: str,s2: int,obs: str) -> float:
		"""
		Returns the probability of moving from state ``s1`` executing `action`
		to ``s2`` generating observation ``obs``.

		Parameters
		----------
		s1: int
			source state ID.
		action: str
			An action.
		s2: int
			destination state ID.
		obs: str
			generated observation.
		
		Returns
		-------
		float
			A probability.
		"""
		return self.h.tau(s1,act,s2,obs)

	def _computeAlphas_MDP(self,sequence: list, sequence_actions: list) -> array:
		"""
		Compute the alpha values for ``sequence`` under the current BW
		hypothesis.

		Parameters
		----------
		sequence : list of str
			sequence of observations.
		sequence_actions : list of str
			sequence of actions.

		Returns
		-------
		2-D narray
			array containing the alpha values.
		"""
		len_seq = len(sequence)
		init_arr = self.h.initial_state
		zero_arr = zeros(shape=((len_seq-1)*self.nb_states,))
		alpha_matrix = append(init_arr,zero_arr).reshape(len_seq,self.nb_states)
		for k in range(len_seq-1):
			for s in range(self.nb_states):
				p = array([self._h_tau(ss,sequence_actions[k],s,sequence[k]) for ss in range(self.nb_states)])
				alpha_matrix[k+1,s] = dot(alpha_matrix[k],p)
		
		alpha_matrix[-1] *= (array(self.h.labelling) == sequence[-1])
		return alpha_matrix.T

	def _computeBetas_MDP(self,sequence: list,sequence_actions: list) -> array:
		"""
		Compute the beta values for ``sequence`` under the current BW
		hypothesis.

		Parameters
		----------
		sequence : list of str
			sequence of observations.
		sequence_actions : list of str
			sequence of actions.

		Returns
		-------
		2-D narray
			array containing the beta values.
		"""
		len_seq = len(sequence)
		init_arr = ones(self.nb_states)*(array(self.h.labelling) == sequence[-1])
		zero_arr = zeros(shape=((len_seq-1)*self.nb_states,))
		beta_matrix = append(zero_arr,init_arr).reshape(len_seq,self.nb_states)
		for k in range(len(sequence)-2,-1,-1):
			for s in range(self.nb_states):
				p = array([self._h_tau(s,sequence_actions[k],ss,sequence[k]) for ss in range(self.nb_states)])
				beta_matrix[k,s] = dot(beta_matrix[k+1],p)
		return beta_matrix.T

	def _processWork_MDP(self,sequence,times):
		sequence_actions = [sequence[i+1] for i in range(0,len(sequence)-1,2)]
		sequence_obs = [sequence[i] for i in range(0,len(sequence)-1,2)]+[sequence[-1]]
		alpha_matrix = self._computeAlphas(sequence_obs,sequence_actions)
		beta_matrix = self._computeBetas(sequence_obs,sequence_actions)
		proba_seq = alpha_matrix.T[-1].sum()
		if proba_seq != 0.0:
			den = zeros(shape=(self.nb_states,len(self.actions)))
			num = zeros(shape=(self.nb_states,len(self.actions),self.nb_states))
			
			for s in range(self.nb_states):
				for i,a in enumerate(self.actions):
					arr_dirak = array([1.0 if t == a else 0.0 for t in sequence_actions])
					den[s,i] += dot(alpha_matrix[s][:-1]*beta_matrix[s][:-1]*arr_dirak,times/proba_seq).sum()
			
				for ss in range(self.nb_states):
					p = array([self._h_tau(s,a,ss,o) for o,a in zip(sequence_obs[:-1],sequence_actions)])
					for ia,act in enumerate(self.actions):
						arr_dirak = [1.0 if a == act else 0.0 for a in sequence_actions]
						num[s,ia,ss] = dot(alpha_matrix[s][:-1]*arr_dirak*beta_matrix[ss][1:]*p,times/proba_seq).sum()
			return [den,num,proba_seq,times]
		return False

	def _update_MDP(self,temp):
		den = array([i[0] for i in temp]).sum(axis=0)
		if type(den) == float64: # den == 0.0
			msg = 'The current hypothesis is not able to generate any of the '
			msg+= 'sequence in the training set.'
			raise ValueError(msg)
		
		num = array([i[1] for i in temp]).sum(axis=0)
		lst_proba=array([i[2] for i in temp])
		lst_times=array([i[3] for i in temp])

		currentloglikelihood = dot(log(lst_proba),lst_times)

		for s in range(self.nb_states):
			for a in range(len(self.actions)):
				if den[s][a] == 0.0 or isnan(den[s][a]):
					den[s][a] = 1.0
					num[s][a] = self.h.matrix[s][a]
		den = den.reshape(self.nb_states,len(self.actions),1)
		matrix = num/den
		self.h.matrix = matrix
		return currentloglikelihood


	# CTMC---------------------------------------------------------------------
	def _h_e(self,s: int) -> float:
		"""
		Returns the exit rate, in the current hypothesis, of state ``s``, i.e.
		the sum of all the rates in this state.

		Parameters
		----------
		s : int
			A state ID.
		
		Returns
		-------
		float
			An exit rate.
		"""
		return self.h.e(s)
	
	def _h_l(self, s1: int, s2: int, obs: str) -> float:
		"""
		Returns the rate, in the current hypothesis, associated to the
		transition from state `s1` to state `s2` generating observation `obs`.

		Parameters
		----------
		s1 : int
			A state ID.
		s2 : int
			A state ID.
		obs : str
			An observation.

		Returns
		-------
		float
			A rate.
		"""
		return self.h.l(s1,s2,obs)	
	
	def _computeAlphas_timed(self,obs_seq: list, times_seq: list) -> array:
		"""
		Compute the beta values for ``obs_seq`` and ``times_seq`` under the
		current BW hypothesis.

		Parameters
		----------
		obs_seq : list of str
			Sequence of observations.
		times_seq : list of float
			Sequence of waiting times.

		Returns
		-------
		2-D narray
			array containing the beta values.
		"""
		len_seq = len(obs_seq)-1
		init_arr = self.h.initial_state
		zero_arr = zeros(shape=(len_seq*self.nb_states,))
		alpha_matrix = append(init_arr,zero_arr).reshape(len_seq+1,self.nb_states).astype(longdouble)
		for k in range(len_seq):
			for s in range(self.nb_states):
				p = array([self._h_l(ss,s,obs_seq[k])*exp(-self._h_e(ss)*times_seq[k]) for ss in range(self.nb_states)])
				alpha_matrix[k+1,s] = dot(alpha_matrix[k],p)
		alpha_matrix[-1] *= (array(self.h.labelling) == obs_seq[-1])
		return alpha_matrix.T

	def _computeBetas_timed(self,obs_seq: list, times_seq: list) -> array:
		"""
		Compute the beta values for ``obs_seq`` and ``times_seq`` under the
		current BW hypothesis.

		Parameters
		----------
		obs_seq : list of str
			Sequence of observations.
		times_seq : list of float
			Sequence of waiting times.

		Returns
		-------
		2-D narray
			array containing the beta values.
		"""
		len_seq = len(obs_seq)-1
		init_arr = ones(self.nb_states)*(array(self.h.labelling) == obs_seq[-1])
		zero_arr = zeros(shape=(len_seq*self.nb_states,))
		beta_matrix = append(zero_arr,init_arr).reshape(len_seq+1,self.nb_states).astype(longdouble)
		for k in range(len_seq-1,-1,-1):
			for s in range(self.nb_states):
				p = array([self._h_l(s,ss,obs_seq[k]) for ss in range(self.nb_states)])
				p = p * exp(-self._h_e(s)*times_seq[k])
				beta_matrix[k,s] = dot(beta_matrix[k+1],p)
		return beta_matrix.T

	def _splitTime(self,sequence: list) -> tuple:
		"""
		Given a trace it returns a sequence of observation and a sequence of
		waiting times. If the given trace is non-timed the output waiting time
		sequence is ``None``.

		Parameters
		----------
		sequence : list
			the trace to split.

		Returns
		-------
		tuple
			a tuple containing a sequence of observations and a sequence of
			waiting times.
		"""
		if type(sequence[1]) == float and type(sequence[0]) == str:
			times_seq = [sequence[i] for i in range(1,len(sequence),2)]
			obs_seq   = [sequence[i] for i in range(0,len(sequence),2)]
		else:
			times_seq = None
			obs_seq = sequence
		return (times_seq,obs_seq)

	def _processWork_CTMC(self,sequence: list, times: int):
		times_seq, obs_seq = self._splitTime(sequence)
		if times_seq == None:
			timed = False
		else:
			timed = True
		alpha_matrix = self._computeAlphas(obs_seq, times_seq)
		beta_matrix  = self._computeBetas( obs_seq, times_seq)
		proba_seq = alpha_matrix.T[-1].sum()
		if proba_seq == 0.0:
			return False
		####################
		if timed:
			den = (alpha_matrix[:,:-1]*beta_matrix[:,:-1]*times_seq*times/proba_seq).sum(axis=1)	
		else:
			den = (alpha_matrix[:,:-1]*beta_matrix[:,:-1]*times/proba_seq).sum(axis=1)

		num = zeros(shape=(self.nb_states,self.nb_states))	
		for s in range(self.nb_states):
			for ss in range(self.nb_states):
				if not self.fixed_parameters[s,ss]:
					if timed:
						p = array([self._h_l(s,ss,o)*exp(-self._h_e(s)*t) for o,t in zip(obs_seq,times_seq)])
					else:
						p = array([self._h_l(s,ss,o)/self._h_e(s) for o in obs_seq]) # not sure
					num[s,ss] = dot(alpha_matrix[s][:-1]*p*beta_matrix[ss][1:],times/proba_seq).sum()
				else:
					num[s,ss] = 0.0
		####################
		return [den, num, proba_seq, times]

	def _update_CTMC(self,temp):
		den = array([i[0] for i in temp]).sum(axis=0)
		num = array([i[1] for i in temp]).sum(axis=0)
		lst_proba=array([i[2] for i in temp])
		lst_times=array([i[3] for i in temp])

		currentloglikelihood = dot(log(lst_proba),lst_times)

		for s in range(len(den)):
			if den[s] == 0.0:
				den[s] = 1.0
				num[s] = self.h.matrix[s]

		matrix = num/den[:, newaxis]
		matrix = self.h.matrix*self.fixed_parameters + matrix

		self.h.matrix = matrix
		return currentloglikelihood


	# HMM----------------------------------------------------------------------
	def _processWork_HMM(self,sequence,times):
		alpha_matrix = self._computeAlphas(sequence)
		beta_matrix = self._computeBetas(sequence)
		proba_seq = alpha_matrix.T[-1].sum()
		if proba_seq != 0.0:
			####################
			den = (alpha_matrix.T[:-1]*beta_matrix.T[:-1]*times/proba_seq).sum(axis=0)
			####################
			num_a = zeros(shape=(self.nb_states,self.nb_states))
			num_b = zeros(shape=(self.nb_states,len(self.alphabet)))
			for s in range(self.nb_states):
				for ss in range(self.nb_states):
					p = array([self._h_tau(s,ss,o) for o in sequence])
					num_a[s,ss] = dot(alpha_matrix[s][:-1]*p*beta_matrix[ss][1:],times/proba_seq).sum()
					temp = alpha_matrix[s][:-1]*beta_matrix[s][:-1]*times/proba_seq
					for o, obs in enumerate(self.alphabet):
						arr_dirak = [1.0 if t == obs else 0.0 for t in sequence]
						num_b[s,o] = (temp*arr_dirak).sum()
			####################
			num_init = alpha_matrix.T[0]*beta_matrix.T[0]*times/proba_seq
			####################
			return [den, num_a, num_b, proba_seq, times, num_init]
		return False

	def _update_HMM(self,temp):
		den = array([i[0] for i in temp]).sum(axis=0)
		num_a = array([i[1] for i in temp]).sum(axis=0)
		num_b = array([i[2] for i in temp]).sum(axis=0)
		lst_proba=array([i[3] for i in temp])
		lst_times=array([i[4] for i in temp])
		lst_init =array([i[5] for i in temp]).T
		
		currentloglikelihood = dot(log(lst_proba),lst_times)

		for s in range(len(den)):
			if den[s] == 0.0:
				den[s] = 1.0
				num_a[s] = self.h.matrix[s]
				num_b[s] = self.h.output[s]

		matrix = num_a/den[:, newaxis]
		output = num_b/den[:, newaxis]

		initial_state = array([lst_init[s].sum()/lst_init.sum() for s in range(self.nb_states)])


		self.h.matrix = matrix
		self.h.output = output
		self.h.initial_state = initial_state
		return currentloglikelihood

	# GoHMM--------------------------------------------------------------------
	def _processWork_GoHMM(self,sequence,times):
		sequence = array(sequence)
		alpha_matrix = self._computeAlphas(sequence,longdouble)
		if isnan(alpha_matrix).any():
			return False
		beta_matrix = self._computeBetas(sequence,longdouble)
		proba_seq = alpha_matrix.T[-1].sum()
		if proba_seq != 0.0:
			den = (alpha_matrix.T[:-1]*beta_matrix.T[:-1]*times/proba_seq).sum(axis=0)
			num_a  = zeros(shape=(self.nb_states,self.nb_states))
			num_mu = zeros(shape=(self.nb_states,self.nb_distr))
			num_va = zeros(shape=(self.nb_states,self.nb_distr))
			for s in range(self.nb_states):
				num_mu[s] = dot(alpha_matrix[s][:-1]*beta_matrix[s][:-1]*sequence.T,times/proba_seq).sum(axis=1)
				num_va[s] = dot(alpha_matrix[s][:-1]*beta_matrix[s][:-1]*(sequence-self.h.mu(s)).T**2,times/proba_seq).sum(axis=1)
				for ss in range(self.nb_states):
					p = array([self._h_tau(s,ss,o) for o in sequence])
					num_a[s,ss] = dot(alpha_matrix[s][:-1]*p*beta_matrix[ss][1:],times/proba_seq).sum()
			num_init = alpha_matrix.T[0]*beta_matrix.T[0]*times/proba_seq
			return [den,num_a,num_mu,num_va,proba_seq,times,num_init]
		return False

	def _update_GoHMM(self,temp):
		a  = array([i[1] for i in temp]).sum(axis=0)
		den= array([i[0] for i in temp]).sum(axis=0)
		mu = array([i[2] for i in temp]).sum(axis=0)
		va = array([i[3] for i in temp]).sum(axis=0)
		lst_proba=array([i[4] for i in temp])
		lst_times=array([i[5] for i in temp])
		init =array([i[6] for i in temp]).sum(axis=0)

		currentloglikelihood = dot(log(lst_proba),lst_times)

		output_mu = (mu.T/den).T
		output_va = sqrt((va.T/den).T)

		matrix = a/den[:, newaxis]
		output = stack((output_mu,output_va),axis=2)

		initial_state = array([init[s]/init.sum() for s in range(self.nb_states)])
		
		self.h.matrix = matrix
		self.h.output = output
		self.h.initial_state = initial_state
		return currentloglikelihood


	# PCTMC--------------------------------------------------------------------
	def _sortParameters(self,fixed_parameters: list):
		"""
		Sort the parameters into the three categories.
		Depending on the category, the parameters are estimated differently:
		in some cases some optimizations are possible (see the paper).
		
		Parameters
		----------
		fixed_parameters : list
			list of all fixed parameters (the one we are not estimating).
		"""
		self.a_pis = zeros((self.h.nb_states,self.h.nb_states,len(self.h.parameter_str)),dtype=float16)
		for iparam,param in enumerate(self.h.parameter_str):
			if not param in fixed_parameters:
				for s,ss in self.h.parameterIndexes(param):
					self.a_pis[s,ss,iparam] = self._a_pi(s,ss,param)

		self.parameters_cat = [[],[],[]]
		for ip,p in enumerate(self.h.parameter_str):
			if not p in fixed_parameters:
				apis = [self.a_pis[x,y,ip] for x,y in self.h.parameterIndexes(p)]
				cs = [self._C(x,y) for x,y in self.h.parameterIndexes(p)]
				if min(apis) == 1 and max(apis) == 1 and min(cs) == 1 and max(cs) == 1:
					self.parameters_cat[0].append(p)
				elif min(cs) == max(cs):
					self.parameters_cat[1].append(p)
				else:
					self.parameters_cat[2].append(p)
		
		self.c_pis = zeros((self.h.nb_states,self.h.nb_states))
		for iparam,param in enumerate(self.parameters_cat[0]):
			for s,ss in self.h.parameterIndexes(param):
				self.c_pis[s,ss] = self._c_pi(s,ss,param)

	def _computeTaus_PCTMC(self):
		self.hval = zeros((self.nb_states,self.nb_states))
		for s in range(self.nb_states):
			for ss in range(self.nb_states):
				self.hval[s,ss] =  self.h.transitionValue(s,ss)

	def _a_pi(self,s1: int,s2: int, p: str) -> int:
		"""
		Return the power of parameter `p` in the transition from `s1` to `s2`.

		Parameters
		----------
		s1 : int
			source state ID.
		s2 : int
			destination state ID.
		p : str
			parameter name.

		Returns
		-------
		int
			Return the power of parameter `p` in the transition from `s1` to `s2`.
		"""
		t = self.h.transitionExpression(s1,s2)
		while not t.is_Pow and not t.is_Symbol:
			flag = False
			for a in t.args:
				if p in [x.name for x in a.free_symbols]:
					t = a
					flag = True
					break
			if not flag:
				print(p,'not in transition',s1,s2)
				input()
		if t.is_Pow:
			return t.args[1]
		else:
			return 1

	def _c_pi(self,s1: int,s2: int, p:str) -> float:
		"""
		Return the coefficient of parameter `p` in the transition from
		`s1` to `s2`.

		Parameters
		----------
		s1 : int
			source state ID.
		s2 : int
			destination state ID.
		p : str
			parameter name.

		Returns
		-------
		float
			the coefficient of parameter `p` in the transition from
			`s1` to `s2`.
		"""
		#used only if p is the only non-fixed parameter
		t = self.h.transitionExpression(s1,s2)
		while not t.is_Mul and not t.is_Symbol:
			flag = False
			for a in t.args:
				if p in [x.name for x in a.free_symbols]:
					t = a
					flag = True
					break
			if not flag:
				print(p,'not in transition',s1,s2)
				input()
		if t.is_Mul:
			res = 1.0
			for a in t.args:
				if a.is_Symbol:
					if not a.name == p:
						res *= self.h.parameterValue(a.name)
				else:
					res *= a
			return float(res)
		else:
			return 1

	def _C(self,s1:int ,s2:int) -> int:
		"""
		Returns the sum of the power of all parameter in the transition
		from `s1` to `s2`.

		Parameters
		----------
		s1 : int
			source state ID.
		s2 : int
			destination state ID.

		Returns
		-------
		int
			the sum of the power of all parameter in the transition
		"""
		r = 0
		for p in self.h.involvedParameters(s1,s2):
			r += self.a_pis[s1,s2,self.h.parameter_str.index(p)]
		return r

	def _h_e_PCTMC(self,s: int) -> float:
		"""
		Returns the exit rate, in the current hypothesis, of state ``s``, i.e.
		the sum of all the rates in this state.

		Parameters
		----------
		s : int
			A state ID.

		Returns
		-------
		float
			An exit rate.
		"""
		return self.hval[s].sum()
	
	def _h_l_PCTMC(self, s1: int, s2: int, obs: str) -> float:
		"""
		Returns the rate, in the current hypothesis, associated to the
		transition from state `s1` to state `s2` generating observation `obs`.

		Parameters
		----------
		s1 : int
			A state ID.
		s2 : int
			A state ID.
		obs : str
			An observation.

		Returns
		-------
		float
			A rate.
		"""
		if self.h.labelling[s1] != obs:
			return 0.0
		return self.hval[s1,s2]

	def _h_tau_PCTMC(self,s1: int,s2: int,obs: str) -> float:
		"""
		Probability of moving from ``s1`` to ``s2`` generating ``obs`` under
		the current hypothesis.

		Parameters
		----------
		s1 : int
			source state ID.
		s2 : int
			destination state ID.
		obs : str
			observation.

		Returns
		-------
		float:
			probability of moving from ``s1`` to ``s2`` generating ``obs``.
		"""
		if self.h.labelling[s1] != obs:
			return 0.0
		e = self._h_e(s1)
		if e == 0.0:
			return inf
		return self.hval[s1,s2]/e

	def _processWork_PCTMC(self,sequence: list, times: int):
		times_seq, obs_seq = self._splitTime(sequence)
		if times_seq == None:
			timed = False
		else:
			timed = True
		alpha_matrix = self._computeAlphas(obs_seq, times_seq)
		beta_matrix  = self._computeBetas( obs_seq, times_seq)
		proba_seq = alpha_matrix.T[-1].sum()
		if proba_seq == 0.0:
			return False
		####################
		if self.update_constant:
			num_cste = array(self.h.transition_expr)
			den_cste = ones(len(num_cste))
			for itrans,trans in enumerate(self.h.transition_expr[1:]):
				if trans.is_real:
					s,ss = where(self.h.matrix == itrans+1)
					s,ss = s[0],ss[0]
					if timed:
						p = array([self._h_l(s,ss,o)*exp(-self._h_e(s)*t) for o,t in zip(obs_seq,times_seq)])
						den_cste[itrans+1] = dot(alpha_matrix[s][:-1]*beta_matrix[s][:-1]*times_seq,times/proba_seq).sum()
					else:
						p = array([self._h_l(s,ss,o)/self._h_e(s) for o in obs_seq])
						den_cste[itrans+1] = dot(alpha_matrix[s][:-1]*beta_matrix[s][:-1],times/proba_seq).sum()
					num_cste[itrans+1] = dot(alpha_matrix[s][:-1]*p*beta_matrix[ss][1:],times/proba_seq).sum()
			
		else:
			num_cste = None
			den_cste = None
			
		#print(sequence, proba_seq)
		num_cat1 = zeros(len(self.parameters_cat[0]))
		den_cat1 = zeros(len(self.parameters_cat[0]))
		for iparam,param in enumerate(self.parameters_cat[0]):
			for s,ss in self.h.parameterIndexes(param):
				if timed:
					p = array([self._h_l(s,ss,o)*exp(-self._h_e(s)*t) for o,t in zip(obs_seq,times_seq)])
				else:
					p = array([self._h_l(s,ss,o)/self._h_e(s) for o in obs_seq])

				if p.sum()>0.0:
					if timed:
						den_cat1[iparam] += dot(alpha_matrix[s][:-1]*beta_matrix[s][:-1]*times_seq,self.c_pis[s,ss]*times/proba_seq).sum()
					else:
						den_cat1[iparam] += dot(alpha_matrix[s][:-1]*beta_matrix[s][:-1],self.c_pis[s,ss]*times/(self._h_e(s)*proba_seq)).sum()
					num_cat1[iparam] += dot(alpha_matrix[s][:-1]*p*beta_matrix[ss][1:],times/proba_seq).sum()

		num_cat2 = zeros(len(self.parameters_cat[1]))
		den_cat2 = zeros(len(self.parameters_cat[1]))
		for iparam,param in enumerate(self.parameters_cat[1]):
			p_index = self.h.parameter_str.index(param)
			for s,ss in self.h.parameterIndexes(param):
				if timed:
					p = array([self._h_l(s,ss,o)*exp(-self._h_e(s)*t) for o,t in zip(obs_seq,times_seq)])
					den_cat2[iparam] += dot(alpha_matrix[s][:-1]*beta_matrix[s][:-1]*times_seq,self.a_pis[s,ss,p_index]*self.hval[s,ss]*times/proba_seq).sum()
				else:
					p = array([self._h_l(s,ss,o)/self._h_e(s) for o in obs_seq])
					den_cat2[iparam] += dot(alpha_matrix[s][:-1]*beta_matrix[s][:-1],self.a_pis[s,ss,p_index]*self.hval[s,ss]*times/(proba_seq*self._h_e(s))).sum()
				num_cat2[iparam] += dot(alpha_matrix[s][:-1]*p*beta_matrix[ss][1:],self.a_pis[s,ss,p_index]*times/proba_seq).sum()
		
		terms_cat3 = []
		for iparam,param in enumerate(self.parameters_cat[2]):
			p_index = self.h.parameter_str.index(param)
			temp = [0.0]
			for s,ss in self.h.parameterIndexes(param):
				p = array([self._h_l(s,ss,o)*exp(-self._h_e(s)*t) for o,t in zip(obs_seq,times_seq)])
				temp[0] -= dot(alpha_matrix[s][:-1]*p*beta_matrix[ss][1:],self.a_pis[s,ss,p_index]*times/proba_seq).sum()
				c = self._C(s,ss)
				for _ in range(1+c-len(temp)):
					temp.append(0.0)
				if timed:
					temp[c] += dot(alpha_matrix[s][:-1]*beta_matrix[s][:-1]*times_seq,self.a_pis[s,ss,p_index]*self.hval[s,ss]*times/proba_seq).sum()/(self.h.parameter_values[param]**c)
				else:
					temp[c] += dot(alpha_matrix[s][:-1]*beta_matrix[s][:-1],self.a_pis[s,ss,p_index]*self.hval[s,ss]*times/proba_seq).sum()/(self._h_e(s)*self.h.parameter_values[param]**c)

			terms_cat3.append(array(temp))

		return [den_cste, num_cste, den_cat1, num_cat1, den_cat2, num_cat2, terms_cat3, proba_seq, times]
			
	def _update_PCTMC(self,temp):
		den_cat1 = array([i[2] for i in temp]).sum(axis=0)
		num_cat1 = array([i[3] for i in temp]).sum(axis=0)
		den_cat2 = array([i[4] for i in temp]).sum(axis=0)
		num_cat2 = array([i[5] for i in temp]).sum(axis=0)
		terms_cat3 = [i[6] for i in temp]
		lst_proba=array([i[7] for i in temp])
		lst_times=array([i[8] for i in temp])

		currentloglikelihood = dot(log(lst_proba),lst_times)

		parameters = self.parameters_cat[0]+self.parameters_cat[1]+self.parameters_cat[2]
		values = []

		if self.update_constant:
			den_cste = array([i[0] for i in temp]).sum(axis=0)
			num_cste = array([i[1] for i in temp]).sum(axis=0)
			for p in range(len(den_cste)):
				if den_cste[p] == 0.0:
					den_cste[p] = 1.0
					num_cste[p] = self.h.transition_expr[p]
			self.h.transition_expr = [sympify(i) for i in (num_cste/den_cste).tolist()]
	
		for p in range(len(den_cat1)):
			if den_cat1[p] == 0.0:
				den_cat1[p] = 1.0
				num_cat1[p] = self.h.parameter_values[self.parameters_cat[0][p]]
		values += (num_cat1/den_cat1).tolist()

		for ip,p in enumerate(self.parameters_cat[1]):
			if den_cat2[ip] == 0.0:
				den_cat2[ip] = 1.0
				num_cat2[ip] = self.h.parameter_values[p]
			else:
				s,ss = self.h.parameterIndexes(p)[0]
				c = self._C(s,ss)
				den_cat2[ip] = den_cat2[ip]**(1/c)
				num_cat2[ip] = self.h.parameter_values[p]*num_cat2[ip]**(1/c)
		values += (num_cat2/den_cat2).tolist()

		for p in range(len(terms_cat3[0])):
			temp = array([i[p] for i in terms_cat3], dtype=float).sum(axis=0)
			values.append(max(polyroots(temp)))

		self.h.instantiate(parameters,values)
		return currentloglikelihood

	def fit_nonInstantiatedParameters(self, traces, initial_model,
			epsilon: float=0.01, max_it: int= inf,
			pp: str='', verbose: bool = True, return_data: bool = False,
			min_val: float = None, max_val: float = None) -> dict:
		"""
		For PCTMC learning only.
		Fits only the non-instantiated parameters in the initial model
		according to ``traces``.

		Parameters
		----------
		traces : Set or list or numpy.ndarray
			training set.
		initial_model : PCTMC
			first hypothesis.
		epsilon : float, optional
			the learning process stops when the difference between the
			loglikelihood of the training set under the two last hypothesis is
			lower than ``epsilon``. The lower this value the better the output,
			but the longer the running time. By default 0.01.
		max_it: int
			Maximal number of iterations. The algorithm will stop after `max_it`
			iterations.
			Default is infinity.
		pp : str, optional
			Will be printed at each iteration. By default ''.
		verbose: bool, optional
			Print or not a small recap at the end of the learning.
			Default is True.
		return_data: bool, optional
			If set to True, a dictionary containing following values will be
			returned alongside the hypothesis once the learning is done.
			'learning_rounds', 'learning_time', 'training_set_loglikelihood'.
			Default is False.
		min_val: float, optional
			Minimal value for the randomly instantiated parameters.
			If not set and if the model has at least two instantiated parameters,
			this value is equal to the parameters with the smallest instantiation.
			If not set and if the model has less than two instantiated parameters,
			this value is equal to 0.1.
		max_val : float, optional
			Maximal value for the randomly instantiated parameters.
			If not set and if the model has at least two instantiated parameters,
			this value is equal to the parameters with the highest instantiation.
			If not set and if the model has less than two instantiated parameters,
			this value is equal to 5.0.		

		Returns
		-------
		dict or list
			Dictionary containing the estimated values for the non-indtantiated
			parameters.
			If `return_data` is set to True, returns a list containing:
			- the dictionary described above,
			- the `returned_data` (see parameter description).
		"""
		if initial_model.model_type != PCTMC_ID:
			raise RuntimeError("The initial model must be a PCTMC")
		to_update = []
		for p in initial_model.parameter_str:
			to_update.append(p)
		
		return self.fit_parameters(traces,initial_model,to_update,epsilon,
								max_it,pp,verbose,return_data,min_val,max_val)

	def fit_parameters(self, traces, initial_model, to_update: list,
			epsilon: float=0.01, max_it: int= inf,
			pp: str='', verbose: bool = True, return_data: bool = False,
			min_val: float = None, max_val: float = None) -> dict:
		"""
		For PCTMC learning only.
		Fits only the non-instantiated parameters in the initial model
		according to ``traces``.

		Parameters
		----------
		traces : Set or list or numpy.ndarray
			training set.
		initial_model : PCTMC
			first hypothesis.
		to_update: list of str
			list of parameter names to update
		epsilon : float, optional
			the learning process stops when the difference between the
			loglikelihood of the training set under the two last hypothesis is
			lower than ``epsilon``. The lower this value the better the output,
			but the longer the running time. By default 0.01.
		max_it: int
			Maximal number of iterations. The algorithm will stop after `max_it`
			iterations.
			Default is infinity.
		pp : str, optional
			Will be printed at each iteration. By default ''.
		verbose: bool, optional
			Print or not a small recap at the end of the learning.
			Default is True.
		return_data: bool, optional
			If set to True, a dictionary containing following values will be
			returned alongside the hypothesis once the learning is done.
			'learning_rounds', 'learning_time', 'training_set_loglikelihood'.
			Default is False.
		min_val: float, optional
			Minimal value for the randomly instantiated parameters.
			If not set and if the model has at least two instantiated parameters,
			this value is equal to the parameters with the smallest instantiation.
			If not set and if the model has less than two instantiated parameters,
			this value is equal to 0.1.
		max_val : float, optional
			Maximal value for the randomly instantiated parameters.
			If not set and if the model has at least two instantiated parameters,
			this value is equal to the parameters with the highest instantiation.
			If not set and if the model has less than two instantiated parameters,
			this value is equal to 5.0.		

		Returns
		-------
		dict or list
			Dictionary containing the estimated values for the non-indtantiated
			parameters.
			If `return_data` is set to True, returns a list containing:
			- the dictionary described above,
			- the `returned_data` (see parameter description).
		"""
		self.type_model = PCTMC_ID
		if initial_model.model_type != PCTMC_ID:
			raise RuntimeError("The initial model must be a PCTMC")
		self.h = initial_model

		update_constant = False
		fixed_parameters = []
		
		for p in initial_model.parameter_str:
			if not p in to_update:
				fixed_parameters.append(p)
		for p in to_update:
			if not p in initial_model.parameter_str:
				raise RuntimeError('Parameter '+p+' is not in the initial model.')

		print("Fitting only parameters: ",end='')
		print(', '.join(to_update))

		# remove all the states never used in the training set:
		# we only care here about the values of some parameters,
		# the structure of the model is out of scope.
		if type(traces) != Set:
			traces = Set(traces)
		alphabet = traces.getAlphabet()
		print("Removing unused states: ", end='')
		print('from',initial_model.nb_states,'states to',end=' ')
		s = 0
		while s < initial_model.nb_states:
			if initial_model.getLabel(s) not in alphabet:
				initial_model.nb_states -= 1
				initial_model.labelling = initial_model.labelling[:s]+initial_model.labelling[s+1:]
				initial_model.initial_state = delete(initial_model.initial_state,s)
				initial_model.matrix = delete(initial_model.matrix,s,0)
				initial_model.matrix = delete(initial_model.matrix,s,1)
				for j in range(len(initial_model.parameter_indexes)):
					k = 0
					while k < len(initial_model.parameter_indexes[j]): 
						if s in initial_model.parameter_indexes[j][k]:
							initial_model.parameter_indexes[j] = initial_model.parameter_indexes[j][:k]+initial_model.parameter_indexes[j][k+1:]
						else:
							if initial_model.parameter_indexes[j][k][0] > s:
								initial_model.parameter_indexes[j][k][0] -= 1
							if initial_model.parameter_indexes[j][k][1] > s:
								initial_model.parameter_indexes[j][k][1] -= 1
							k += 1
				s = -1
			s += 1
		t = 1
		l = len(initial_model.transition_expr)
		while t < l:
			if not (initial_model.matrix == t).any():
				initial_model.transition_expr = initial_model.transition_expr[:t]+initial_model.transition_expr[t+1:]
				initial_model.matrix -= (initial_model.matrix > t)
				l -= 1
			else:
				t += 1
		#initial_model.alphabet = alphabet
		print(initial_model.nb_states)

		self._getTrainingSet(traces)
		
		if min_val != None:
			if max_val != None:
				self.h.randomInstantiation(min_val=min_val, max_val=max_val)
			else:
				self.h.randomInstantiation(min_val=min_val)
		elif max_val != None:
			self.h.randomInstantiation(max_val=max_val)
		
		if self.training_set.type == 4:
			self._computeAlphas = self._computeAlphas_timed
			self._computeBetas  = self._computeBetas_timed
		self.nb_parameters = self.h.nb_parameters
		self.update_constant = update_constant
		self._processWork = self._processWork_PCTMC
		self._update = self._update_PCTMC
		self._computeTaus = self._computeTaus_PCTMC
		self._h_e = self._h_e_PCTMC
		self._h_l = self._h_l_PCTMC
		self._h_tau=self._h_tau_PCTMC
		self._sortParameters(fixed_parameters)

		returned = self._bw(max_it,pp,epsilon,False,False,verbose,False,return_data)
		
		res = {}
		for p in to_update:
			res[p] = self.h.parameterValue(p)
		if return_data:
			return res, returned[1]
		else:
			return res


	# General------------------------------------------------------------------
	def _getInitialModel(self,initial_model, nb_states: int):

		if initial_model == None:
			if nb_states == None:
				raise RuntimeError("initial_state or the nb_states must be given. Here none of them is.")
			else:
				self.h = None
				self.type_model = None
				return	
		try:
			initial_model.name
		except AttributeError: # then initial_model is a stormpy sparse model
			if not self.stormpy_installed:
				raise RuntimeError("the initial model is a Storm model and Storm is not installed on the machine")
			initial_model = self.stormpyModeltoJajapy(initial_model)
		self.h = initial_model
		self.type_model = initial_model.model_type
			
	def _getTrainingSet(self,training_set):
		if type(training_set) == ndarray:
			training_set = training_set.tolist()
		if type(training_set) == list:
			training_set = Set(training_set)
			if training_set.type == 0:
				a = set()
				l = set()
				mdp = True
				for seq in training_set.sequences:
					a = a.union(set([seq[k] for k in range(1,len(seq),2)]))
					l = l.union(set([seq[k] for k in range(0,len(seq),2)]))
					if a.intersection(l) != {}:
						mdp = False
						break
				if mdp:
					training_set.type = MDP_ID
		if training_set.type != 3 and self.type_model != HMM_ID:
			if training_set.type == 1:
				_, self.alphabet = training_set.getActionsObservations()
			else:
				self.alphabet = training_set.getAlphabet()
			if not 'init' in self.alphabet:
				for s in range(len(training_set.sequences)):
					training_set.sequences[s].insert(0,'init')
			else:
				self.alphabet.remove('init')

		self.training_set = training_set

	def _createInitialModel(self,nb_states, random_initial_state,
							min_exit_rate_time, max_exit_rate_time, self_loop,
							nb_distributions):
		if self.training_set.type == 0:
			from ..mc.MC import MC_random
			self.type_model = MC_ID
			self.h = MC_random(nb_states,self.alphabet,random_initial_state)
		
		elif self.training_set.type == 1:
			from ..mdp.MDP import MDP_random
			self.type_model = MDP_ID
			act,_ = self.training_set.getActionsObservations()
			self.h = MDP_random(nb_states,self.alphabet,act,random_initial_state)
		
		elif self.training_set.type == 3:
			from ..gohmm.GoHMM import GoHMM_random
			if nb_distributions == None:
				nb_distributions = len(self.training_set.sequences[0][0])
			self.type_model = GOHMM_ID
			self.h = GoHMM_random(nb_states,nb_distributions,random_initial_state)
		
		elif self.training_set.type == 4:
			from ..ctmc.CTMC import CTMC_random
			if min_exit_rate_time == None or max_exit_rate_time == None:
				raise RuntimeError("The minimum and maximum exit rate must be given if the initial hypothesis is not provided.")
			self.type_model = CTMC_ID
			self.h = CTMC_random(nb_states,self.alphabet,min_exit_rate_time,max_exit_rate_time,self_loop,random_initial_state)

	def _computeAlphas(self, sequence: list, dtype=False) -> array:
		"""
		Compute the alpha values for ``sequence`` under the current BW
		hypothesis.

		Parameters
		----------
		sequence : list of str
			sequence of observations.
		dtype : numpy.scalar
			If it set, the output will be a numpy array of this type,
			otherwise it is a numpy array of float64.

		Returns
		-------
		2-D narray
			array containing the alpha values.
		"""
		len_seq = len(sequence)
		init_arr = self.h.initial_state
		zero_arr = zeros(shape=(len_seq*self.nb_states,))
		alpha_matrix = append(init_arr,zero_arr).reshape(len_seq+1,self.nb_states)
		if dtype != False:
			alpha_matrix = alpha_matrix.astype(dtype)
		else:
			dtype = float64
		for k in range(len_seq):
			for s in range(self.nb_states):
				p = array([self._h_tau(ss,s,sequence[k]) for ss in range(self.nb_states)],dtype=dtype)
				alpha_matrix[k+1,s] = dot(alpha_matrix[k],p)
		return alpha_matrix.T

	def _computeBetas(self, sequence: list, dtype=False) -> array:
		"""
		Compute the beta values for ``sequence`` under the current BW
		hypothesis.

		Parameters
		----------
		sequence : list of str
			sequence of observations.
		dtype : numpy.scalar
			If it set, the output will be a numpy array of this type,
			otherwise it is a numpy array of float64.

		Returns
		-------
		2-D narray
			array containing the beta values.
		"""
		len_seq = len(sequence)
		init_arr = ones(self.nb_states)
		zero_arr = zeros(shape=(len_seq*self.nb_states,))
		beta_matrix = append(zero_arr,init_arr).reshape(len_seq+1,self.nb_states)
		if dtype != False:
			beta_matrix = beta_matrix.astype(dtype)
		else:
			dtype = float64
		for k in range(len(sequence)-1,-1,-1):
			for s in range(self.nb_states):
				p = array([self._h_tau(s,ss,sequence[k]) for ss in range(self.nb_states)],dtype=dtype)
				beta_matrix[k,s] = dot(beta_matrix[k+1],p)
		return beta_matrix.T

	def _computeTaus(self):
		pass
	
	def _runProcesses(self,training_set):
		if platform != "win32" and platform != "darwin" and NB_PROCESS > 1:
			p = Pool(processes = NB_PROCESS)
			tasks = []
			for seq,times in zip(training_set.sequences,training_set.times):
				tasks.append(p.apply_async(self._processWork, [seq, times,]))
			return [res.get() for res in tasks if res.get() != False]
		else:
			return [self._processWork(seq, times) for seq,times in zip(training_set.sequences,training_set.times)]
	
	def _endPrint(self,it,rt):
		print()
		print("---------------------------------------------")
		print("Learning finished")
		print("Iterations:\t  ",it)
		print("Running time:\t  ",rt)
		print("---------------------------------------------")
		print()

	def _h_tau(self,s1: int,s2: int,obs: str) -> float:
		"""
		Probability of moving from ``s1`` to ``s2`` generating ``obs`` under
		the current hypothesis.

		Parameters
		----------
		s1 : int
			source state ID.
		s2 : int
			destination state ID.
		obs : str
			observation.

		Returns
		-------
		float:
			probability of moving from ``s1`` to ``s2`` generating ``obs``.
		"""
		return self.h.tau(s1,s2,obs)



import unittest
from ..mc import *
from os import remove
from ..base.Set import *
from ..base.BW import BW
from math import log
from numpy import where, array

def modelMC_REBER():
	labelling = list("BTSXSPTXPVVE")
	initial_state = 0
	name = "MC_REBER"
	transitions = [(0,1,0.5),(0,5,0.5),(1,2,0.6),(1,3,0.4),(2,2,0.6),(2,3,0.4),
				   (3,7,0.5),(3,4,0.5),(4,11,1.0),(5,6,0.7),(5,9,0.3),
				   (6,6,0.7),(6,9,0.3),(7,6,0.7),(7,9,0.3),(8,7,0.5),(8,4,0.5),
				   (9,8,0.5),(9,10,0.5),(10,11,1.0),(11,11,1.0)]
	return createMC(transitions,labelling,initial_state,name)

m = modelMC_REBER()

class MCTestclass(unittest.TestCase):

	def test_MC_initial_state(var):
		labelling=['a','b','c','d','a']
		transitions = [(0,1,0.8),(0,2,0.2),
				   (1,3,0.6),(1,2,0.4),
				   (2,0,0.5),(2,4,0.5),
				   (3,2,0.3),(3,3,0.7),
				   (4,2,0.2),(4,3,0.1),(4,4,0.7)]
		mc = createMC(transitions,labelling,0)
		var.assertEqual(mc.nb_states,6)
		var.assertEqual(mc.labelling.count('init'),1)
		var.assertEqual(mc.getLabel(int(where(mc.initial_state == 1.0)[0][0])),'init')
		
		labelling=['a','b','c','d','a']
		mc = createMC(transitions,labelling,[0.3,0.0,0.0,0.2,0.5])
		var.assertEqual(mc.nb_states,6)
		var.assertEqual(mc.labelling.count('init'),1)
		var.assertEqual(mc.pi(5),1.0)
		var.assertTrue((mc.matrix[-1]==array([0.3,0.0,0.0,0.2,0.5,0.0])).all())
		
		labelling=['a','b','c','d','a']
		mc = createMC(transitions,labelling,array([0.3,0.0,0.0,0.2,0.5]))
		var.assertEqual(mc.nb_states,6)
		var.assertEqual(mc.labelling.count('init'),1)
		var.assertEqual(mc.pi(5),1.0)
		var.assertTrue((mc.matrix[-1]==array([0.3,0.0,0.0,0.2,0.5,0.0])).all())
		

	def test_MC_state(var):
		var.assertEqual(m.tau(1,0,'B'),0.0)
		var.assertEqual(m.tau(0,1,'B'),0.5)
		var.assertEqual(m.getLabel(2),'S')
		var.assertEqual(m.tau(2,3,'S'),0.4)
		var.assertEqual(m.tau(2,3,'something else'),0.0)
		var.assertEqual(m.getLabel(1),'T')

	
	def test_MC_save_load_str(var):
		m.save("test_save.txt")
		mprime = loadMC("test_save.txt")
		var.assertEqual(str(m),str(mprime))
		remove("test_save.txt")
	
	def test_MC_getAlphabet(var):
		var.assertEqual(set(m.getAlphabet()),
						set(list("BTPXSVE")+['init']))
	
	def test_MC_Set(var):
		set1 = m.generateSet(50,10)
		set2 = m.generateSet(50,1/4,"geo",6)
		set1.addSet(set2)
		var.assertEqual(set1.type,0)
		set1.save("test_save.txt")
		set2 = loadSet("test_save.txt")
		var.assertTrue(set1.isEqual(set2))
		remove("test_save.txt")

	def test_MC_logLikelihood(var):
		set1 = Set([['init','B','T','X','X']],[1])
		set2 = Set([['init','B','P','T','T']],[2])
		l11 = m._logLikelihood_oneproc(set1)
		l12 = m._logLikelihood_multiproc(set1)
		var.assertAlmostEqual(l11,l12)
		var.assertAlmostEqual(l11,log(0.1))
		l2 = m.logLikelihood(set2)
		var.assertAlmostEqual(l2,log(0.245))
		set1.addSet(set2)
		l3 = m.logLikelihood(set1)
		var.assertAlmostEqual(l3,(log(0.1)+2*log(0.245))/3)
	
	def test_MC_random(var):
		alphabet = list("BTSXPVE")
		random_model = MC_random(11, alphabet, False)
		for i in alphabet:
			var.assertGreaterEqual(random_model.labelling.count(i),1)

	def test_BW_MC(var):
		initial_model   = loadMC("jajapy/tests/materials/mc/random_MC.txt")
		training_set    = loadSet("jajapy/tests/materials/mc/training_set_MC.txt")
		output_expected = loadMC("jajapy/tests/materials/mc/output_MC.txt")
		output_gotten   = BW().fit( training_set, initial_model, stormpy_output=False)
		test_set = m.generateSet(10000,10)
		var.assertAlmostEqual(output_expected.logLikelihood(test_set),
							  output_gotten.logLikelihood(test_set))
	
	def test_Alergia(var):
		training_set    = loadSet("jajapy/tests/materials/mc/training_set_MC.txt")
		output_expected = loadMC("jajapy/tests/materials/mc/output_alergia_MC.txt")
		output_gotten =  Alergia().fit(training_set,0.000005,stormpy_output=False)
		test_set = m.generateSet(10000,10)
		var.assertAlmostEqual(output_expected.logLikelihood(test_set),
							  output_gotten.logLikelihood(test_set))


if __name__ == "__main__":
	unittest.main()


from .MDP import MDP, createMDP, MDP_random, loadMDP
from .Scheduler import UniformScheduler, MemorylessScheduler, FiniteMemoryScheduler
from .IOAlergia import IOAlergia
from .Active_BW_MDP import Active_BW_MDP


from random import random
from scipy.stats import norm
from numpy import nditer, array, ndarray, nonzero
from numpy.random import randint
from math import log,sqrt

def normpdf(x: float, params: list, variation:float = 0.01) -> float:
	"""
	Returns the probability of ``x`` under a normal distribution returns of 
	parameters ``params``. Since this probability should be 0 it returns in
	fact the probability that the normal distribution gives us a value
	between ``x-variation`` and ``x+variation``.

	Parameters
	----------
	x: float
		the value of the normal distribution.
	params: list of two float.
		the parameters of the distribution: [mean,sd].
	variation: float
		the vicinity.

	Returns
	-------
	float
		the probability of ``x`` under a normal distribution returns of
		parameters ``params``.
	"""
	return norm.cdf(x+variation,params[0],params[1]) - norm.cdf(x-variation,params[0],params[1])


def resolveRandom(m: list) -> int:
	"""
	Given a list of probabilities it returns the index of the one choosen
	according to the probabilities.
	Example: if m=[0.7,0.3], it will returns 0 with probability 0.7,
	and 1 with probability 0.3.
	
	Parameters
	----------
	m: list of float
		list of probabilities.
	
	Returns
	-------
	int
		the chosen index.
	
	Examples
	--------
	>>> import jajapy as ja
	>>> p = [0.2,0.6,0.15,0.05]
	>>> chosen = ja.resolveRandom(p)
	>>> print(chosen)
	1
	"""
	m = array(m).cumsum()
	mi = nonzero(m)[0]
	r = random()
	i = 0
	while r > m[mi[i]]:
		i += 1
	
	return mi[i]

def normalize(ll):
	"""
	Normalizes a list of float such that the sum is equal to 1.0.

	Parameters
	----------
	ll: list of float or 1-D narray
		list of probabilities to normalize.

	Returns
	-------
	list or 1-D narray
		normalized list.
	"""
	if type(ll) == list:
		return [i/sum(ll) for i in ll]
	else:
		return array([i/ll.sum() for i in nditer(ll)])

def randomProbabilities(size: int) -> ndarray:
	"""
	Return of ndarray of length ``size`` of probabilities.

	Parameters
	----------
	size: int
		size of the output list.

	Returns
	-------
	list of float
		list of probabilities.
	
	Raises
	------
	ValueError
		If `size` is strictly lower than 1.0.
	TypeError
		If `size` is not an `int`.
	
	Examples
	--------
	>>> import jajapy as ja
	>>> p = ja.randomProbabilities(4)
	>>> print(p)
	array([0.3155861636575178, 0.11453783121165262, 0.5686125794652406, 0.001263425665589013])
	"""
	if size <= 0:
		raise ValueError("The size parameter should be higher than 0.")
	if type(size) != int:
		raise TypeError("The size parameter should be an int.")
	return normalize(randint(1,11,size))

def checkProbabilities(l: ndarray) -> bool:
	"""
	Check if a ndarray contains probabilities i.e. check if the sum of the
	ndarray is 0.0 or 1.0. 

	Parameters
	----------
	l : ndarray
		List of probabilities

	Returns
	-------
	bool
		True if the sum of the list is 0.0 or 1.0.
	"""
	return round(l.sum(),3) == 1.0 or round(l.sum(),3) == 0.0

def hoeffdingBound(f1,n1,f2,n2,alpha):
	if n1*n2 == 0.0:
		return True
	return abs(f1/n1 - f2/n2) < (n1**(-1/2)+n2**(-1/2))*sqrt(0.5*log(2/alpha))


from ..base.tools import resolveRandom, randomProbabilities, checkProbabilities
from ..base.Base_HMM import Base_HMM,HMM_ID
from ast import literal_eval
from numpy import array, where, zeros

class HMM(Base_HMM):

	def __init__(self,matrix, output, alphabet, initial_state,name="unknown_HMM"):
		"""
		Creates an HMM.

		Parameters
		----------
		matrix : ndarray
			Represents the transition matrix.
			`matrix[s1][s2]` is the probability of moving from `s1` to `s2`.
		output : ndarray or list
			Represents the output matrix.
			`output[s1][obs]` is the probability of seeing `alphabet[obs]` in 
			state `s1`. 
		alphabet: list
			The list of all possible alphabet, such that:
			`alphabet.index("obs")` is the ID of `obs`.
		initial_state : int or list of float
			Determine which state is the initial one (then it's the id of the
			state), or what are the probability to start in each state (then it's
			a list of probabilities).
		name : str, optional
			Name of the model.
			Default is "unknow_MC"
		"""
		self.model_type = HMM_ID
		self.alphabet = alphabet
		self.output = array(output)
		super().__init__(matrix,initial_state,name)
		for i in range(self.nb_states):
			if not checkProbabilities(output[i]):
				raise ValueError("The probability to generate an observation in state",i,"should be 1.0, here it's",output[i].sum())

	def b(self, s: int, l: str) -> float:
		"""
		Returns the probability of generating `l` in state `s`.
		If `s` is not a valid state ID it returns 0.

		Parameters
		----------
		s : int
			ID of the source state.		
		l : str
			observation.
		
		Returns
		-------
		float
			probability of generating `l` in state `s`.

		Examples
		--------
		>>> model.b(0,'x')
		0.4
		>>> model.b(0,'foo')
		0.0
		"""
		if s < 0 or s >= self.nb_states or l not in self.alphabet:
			return 0.0 
		return self.output[s][self.alphabet.index(l)]
	
	def tau(self, s1: int, s2: int, obs: str) -> float:
		"""
		Return the probability of generating from state `s1` observation `obs` and moving to state `s2`.
		If `s1` or `s2` is not a valid state ID it returns 0.

		Parameters
		----------
		s1 : int
			ID of the source state.
		s2 : int
			ID of the destination state.
		obs : str
			An observation.

		Returns
		-------
		float
			The probability of generating from state `s1` observation `obs` and moving to state `s`.
		
		Examples
		--------
		>>> model.tau(0,1,'x')
		0.2
		"""
		return super().tau(s1,s2,obs)
	
	def getAlphabet(self,state: int = -1) -> list:
		"""
		If state is set, returns the list of all the observations we could
		see in `state`. Otherwise it returns the alphabet of the model. 

		Parameters
		----------
		state : int, optional
			a state ID

		Returns
		-------
		list of str
			list of observations
		"""
		if state == -1:
			return self.alphabet
		else:
			self._checkStateIndex(state)
			return [self.alphabet[i] for i in where(self.output[state] > 0.0)[0]]

	def _stateToString(self, state: int) -> str:
		res = super()._stateToString(state)
		for j in range(len(self.output[state])):
			if self.output[state][j] > 0.0001:
				res += "s"+str(state)+" => "+str(self.alphabet[j])+" : "+str(self.output[state][j])+'\n'
		return res
	
	def save(self,file_path:str):
		"""
		Save the model into a text file.

		Parameters
		----------
		file_path : str
			path of the output file.
		
		Examples
		--------
		>>> model.save("my_model.txt")
		"""
		f = open(file_path, 'w')
		f.write("HMM\n")
		f.write(str(self.alphabet))
		f.write('\n')
		super()._save(f)

	def next_obs(self, state: int) -> str:
		"""
		Generates one observation according to the distribution described by `self.output_matrix`.
		
		Returns
		-------
		str
			An observation.
		"""
		c = resolveRandom(self.output[state].flatten())
		return self.alphabet[c]


def loadHMM(file_path: str) -> HMM:
	"""
	Load an HMM saved into a text file.

	Parameters
	----------
	file_path : str
		Location of the text file.
	
	Returns
	-------
	output : HMM
		The HMM saved in `file_path`.
	"""
	f = open(file_path,'r')
	l = f.readline()[:-1] 
	if l != "HMM":
		print("ERROR: this file doesn't describe an HMM: it describes a "+l)
	alphabet = literal_eval(f.readline()[:-1])
	output = literal_eval(f.readline()[:-1])
	output = array(output)
	name = f.readline()[:-1]
	initial_state = array(literal_eval(f.readline()[:-1]))
	matrix = literal_eval(f.readline()[:-1])
	matrix = array(matrix)
	f.close()
	return HMM(matrix, output, alphabet, initial_state, name)


def HMM_random(nb_states: int, alphabet: list, random_initial_state: bool = False) -> HMM:
	"""
	Generates a random HMM.

	Parameters
	----------
	nb_states : int
		Number of states.
	alphabet : list of str
		List of observations.
	random_initial_state: bool, optional
		If set to True we will start in each state with a random probability, otherwise we will always start in state 0.
		Default is False.
	
	Returns
	-------
	HMM
		A pseudo-randomly generated HMM.
	
	Examples
	--------
	>>> m = ja.HMM_random(4,['a','b','x','y'])
	"""
	matrix = []
	output = []
	for s in range(nb_states):
		p1 = array(randomProbabilities(nb_states))
		matrix.append(p1)
		p2 = array(randomProbabilities(len(alphabet)))
		output.append(p2)
	matrix = array(matrix)
	output = array(output)

	if random_initial_state:
		init = randomProbabilities(nb_states)
	else:
		init = 0
	return HMM(matrix, output, alphabet, init,"HMM_random_"+str(nb_states)+"_states")


def createHMM(transitions: list, emission: list, initial_state, name: str ="unknown_HMM") -> HMM:
	"""
	An user-friendly way to create a HMM.

	Parameters
	----------
	transitions : [ list of tuples (int, int, float)]
		Each tuple represents a transition as follow: 
		(source state ID, destination state ID, probability).
	emission : [ list of tuples (int, str, float)]
		Each tuple represents an emission probability as follow: 
		(source state ID, emitted label, probability).
	initial_state : int or list of float
		Determine which state is the initial one (then it's the id of the
		state), or what are the probability to start in each state (then it's
		a list of probabilities).
	name : str, optional
		Name of the model.
		Default is "unknow_HMM"
	
	Returns
	-------
	HMM
		the HMM describes by `transitions`, `emission`, and `initial_state`.
	
	Examples
	--------
	>>> model = createHMM([(0,1,1.0),(1,0,0.6),(1,1,0.4)],[(0,'a',0.8),(0,'b',0.2),(1,'b',1.0)],0,"My_HMM")
	>>> print(model)
	Name: My_HMM
	Initial state: s0
	----STATE s0----
	s0 -> s1 : 1.0
	************
	s0 => a : 0.8
	s0 => b : 0.2	
	----STATE s1----
	s1 -> s0 : 0.6
	s1 -> s1 : 0.4
	************
	s1 => b : 1.0
	"""
	
	states = list(set([i[0] for i in transitions]+[i[1] for i in transitions]))
	states.sort()
	alphabet = list(set([i[1] for i in emission]))
	nb_states = len(states)
	matrix = zeros((nb_states,nb_states))
	for t in transitions:
		matrix[states.index(t[0])][states.index(t[1])] = t[2]
	output = zeros((nb_states,len(alphabet)))
	for t in emission:
		output[states.index(t[0])][alphabet.index(t[1])] = t[2]
	return HMM(matrix, output, alphabet, initial_state, name)



from .MDP import *
from math import sqrt, log

class IOFPTA_state:
	"""
	Class for a IOFPTA state
	"""
	def __init__(self, obs: str, c: int, lbl: list) -> None:
		"""
		Creates a IOFTPA state. 

		Parameters
		----------
		obs : str
			The observation associated to this state.
		c : int
			A counter corresponding to the number of times we went through this
			state.
		lbl : list
			A list of alternating action-observation.
			The prefix associated to this state.
		"""
		self.observation = obs
		self.counter = c
		self.label = lbl
		self.id = -1
		self.transitions = {}

	def counterAdd(self,n:int) -> None:
		"""
		In creases the counter by ``n``

		Parameters
		----------
		n : int
			An integer 
		"""
		self.counter += n

	def setId(self,i:int) -> None:
		"""
		Set the ID of this state to ``i``.

		Parameters
		----------
		i : int
			A new state ID.
		"""
		self.id = i

	def successors(self) -> list:
		"""
		Returns the list of all states that can be reached from this state.

		Returns
		-------
		list
			List of state IDs.
		"""
		res = []
		for a in self.transitions:
			for s in self.transitions[a][1]:
				if s not in res:
					res.append(s)
		res.sort()
		return res

	def successorsAction(self,action: str) -> list:
		"""
		Returns the list of all states that can be reached from this state
		executing ``action``.

		Returns
		-------
		list
			List of state IDs.
		"""
		if not self.actionAllowed(action):
			return []
		return self.transitions[action][1]


	def transitionsAdd(self, action: str, prob: float, state: int) -> None:
		"""
		Add a transition leaving this state.

		Parameters
		----------
		action : str
			An action.
		prob : float
			A probability.
		state : int
			A state ID (the destination state).
		"""
		if not self.actionAllowed(action):
			self.transitions[action] = [[prob], [state]]
		else:
			self.transitions[action][0].append(prob)
			self.transitions[action][1].append(state)

	def transitionChange(self, action: str, old_state: int, new_state: int) -> None:
		"""
		Change the destination state of a given transition.

		Parameters
		----------
		action : str
			An action.
		old_state : int
			A state ID (the previous destination).
		new_state : int
			A state ID (the new destination).
		"""
		self.transitions[action][1][self.transitions[action][1].index(old_state)] = new_state

	def getTransitionProb(self,action: str,s2: int) -> float:
		"""
		Returns the probability that we reach ``s2`` from this state executing
		``action``.

		Parameters
		----------
		action : str
			An action.
		s2 : int
			A state ID.

		Returns
		-------
		float
			A probability.
		"""
		return self.transitions[action][0][self.transitions[action][1].index(s2)]

	def actionAllowed(self,act: str) -> bool:
		"""
		Checks if action ``act`` can be executed in this state.

		Parameters
		----------
		act : str
			An action.

		Returns
		-------
		bool
			True if action ``act`` can be executed in this state.
		"""
		return act in self.transitions

	def actionsAllowed(self) -> list:
		"""
		Returns the list of all actions that can be executed in this state.

		Returns
		-------
		list
			A list of actions.
		"""
		return [a for a in self.transitions]

	def __str__(self) -> str:
		return self.label+'\t'+self.transitions



class IOFPTA:
	"""
	Class for a Input and Output Frequency Prefix Tree Acceptor.
	"""
	def __init__(self,states: list, o: list, a: list) -> None:
		"""
		Creates a IOFPTA

		Parameters
		----------
		states : list of IOFPTA_states
			A list of all the states in this IOFPTA
		o : list of str
			A list of all possible observations
		a : list of str
			A list of all possible actions
		"""
		self.states = states
		self.observations = o
		self.actions = a
		self.initial = []

	def __str__(self) -> str:
		res = ""
		for s in self.states:
			res += str(s)
		return res
	
	def addInitialState(self,state:int) -> None:
		if not state in self.initial:
			self.initial.append(state)

	def successor(self, state: int, action: str, obs: str) -> int:
		"""
		Returns the index of the state we reach after executing ``action`` in
		``state`` and seeing ``obs``.

		Parameters
		----------
		state : int
			A state ID.
		action : str
			An action.
		obs : str
			An observation.

		Returns
		-------
		int
			A state ID.
		"""
		if not self.states[state].actionAllowed(action):
			return None

		for i in self.states[state].successorsAction(action):
			if self.states[i].observation == obs:
				return i

	def compatible(self, s1: int, s2: int, epsilon: float) -> bool:
		"""
		Checks if state `s1` and `s2` are ``epsilon`` compatible.

		Parameters
		----------
		s1 : int
			A state ID.
		s2 : int
			A state ID.
		epsilon : float
			A float between 0 and 1 (0 excluded).

		Returns
		-------
		bool
			True if s1 and s2 are epsilon compatible.
		"""
		if s1 == None or s2 == None:
			return True
		
		if self.states[s1].observation != self.states[s2].observation:
			return False

		for a in self.actions:
			for o in self.observations:
				if not self.hoeffding(s1, s2, a, o, epsilon):
					return False
				if not self.compatible(self.successor(s1, a, o), self.successor(s2, a, o), epsilon):
					return False
		return True
				
	def hoeffding(self, s1: int, s2: int, a: str, o: str, epsilon: float) -> bool:
		"""
		Checks if the distance of distributions of state ``s1`` and state
		``s2`` for action ``a`` and observation ``o`` is within the Hoeffding
		bound with parameter ``epsilon``. 

		Parameters
		----------
		s1 : int
			A state ID.
		s2 : int
			A state ID.
		a : str
			An action.
		o : str
			An observation.
		epsilon : float
			A float between 0 and 1 (0 excluded).			

		Returns
		-------
		bool
			True if the distance of distributions are in the Hoeffding bound.
		"""
		i1 = self.successor(s1, a, o)
		i2 = self.successor(s2, a, o)

		if i1 == None or i2 == None:
			return True
				
		f1 = self.states[s1].getTransitionProb(a, i1)
		n1 = sum(self.states[s1].transitions[a][0])
		f2 = self.states[s2].getTransitionProb(a, i2)
		n2 = sum(self.states[s2].transitions[a][0])

		if n1*n2 == 0:
			return True

		return abs((f1/n1)-(f2/n2)) < (sqrt(1/n1)+sqrt(1/n2))*sqrt(log(2/epsilon)/2)

	def findPredec(self,s: int) -> list:
		"""
		Returns a list of pairs state-action (s',a) s.t., by executing action a
		in state s', we can reach state ``s``.

		Parameters
		----------
		s : int
			A state ID.

		Returns
		-------
		list
			A list of pairs stateID-action.
		"""
		res = []
		for i in range(len(self.states)):
			for a in self.states[i].actionsAllowed():
				for j in self.states[i].successorsAction(a):
					if j == s:
						res.append((i,a))
		return res

	def merge(self,s1:int ,s2: int) -> None:
		"""
		Merges state ``s2`` into state ``s1``.

		Parameters
		----------
		s1 : int
			A state ID.
		s2 : int
			A state ID.
		"""
		#action = self.states[s2].label[-2]
		if s1 == None or s2 == None:
			return None

		predec = self.findPredec(s2)
		for sa in predec:
			self.states[sa[0]].transitionChange(sa[1],s2,s1)

		self.states[s1].counterAdd(self.states[s2].counter)
		self.states[s2].counter = 0 #useless but meaningfull

		for a in self.states[s2].actionsAllowed():
			for o in self.observations:
				succ2 = self.successor(s2,a,o)
				if succ2 != None:
					succ1 = self.successor(s1,a,o)
					if succ1 != None:
						self.merge(succ1,succ2)
						#self.states[s1].transition_prob_add(a,succ1,self.states[s2].getTransitionProb(a,succ2))
						#self.states[succ1].counterAdd(self.states[succ2].counter)
						#self.states[succ2].counter = 0 #useless but meaningfull
					else:
						self.states[s1].transitionsAdd(a,self.states[s2].getTransitionProb(a,succ2),succ2)

	def runSeq(self,seq:list) -> bool:
		"""
		Checks if the IOFTPA can generate the given trace ``seq``.

		Parameters
		----------
		seq : list
			A trace, i.e. a list of alternating action-observation.

		Returns
		-------
		bool
			True if the IOFTPA can generate the given trace ``seq``.
		"""
		for s_current in self.initial:
			i_current = 0
			path = [s_current]
			f = True
			while i_current < len(seq) and f:
				f = False
				for s in self.states[s_current].successorsAction(seq[i_current]):
					if self.states[s].observation == seq[i_current+1]:
						s_current = s
						path.append(s_current)
						i_current += 2
						f = True
						break
				
			if f:
				return True
		return False


	def cleanMDP(self,red: list) -> MDP:
		"""
		Returns a MDP built from a subset of states of this IOFTPA.

		Parameters
		----------
		red : list
			A list of state IDs.

		Returns
		-------
		MDP
			An MDP.
		"""

		actions = self.actions
		states = red
		labelling = [self.states[i].observation for i in states]+['init']
		transitions = zeros((len(states)+1,len(actions),len(states)+1))
		
		for si,s in enumerate([self.states[i] for i in states]):
			for ai,a in enumerate([a for a in actions if s.actionAllowed(a)]):
				tot = sum(s.transitions[a][0])
				for p,s2 in zip(s.transitions[a][0],s.transitions[a][1]):
					transitions[si][ai][states.index(s2)] = p/tot
		transitions[-1][0][0] = 1.0
		return MDP(transitions, labelling, actions)



class IOAlergia:
	"""
	Class for an  IOAlergia algorithm described here:
	https://arxiv.org/pdf/1212.3873.pdf
	"""
	def __init__(self):
		None
		
	def _initialize(self,sample,alpha,actions,observations):
		pass
		self.alpha = alpha
		self.sample = sample

		self.actions = actions
		self.observations = observations

		self.actions.sort()
		self.observations.sort()
		self.N = sum(sample.times)
		self.n = len(sample.sequences[0])
		self.t = self._buildIOFPTA()
		self.a = self.t
		
	def _buildIOFPTA(self):
		states_lbl = []
		states = []
		#init states_lbl and states_counter
		for seq,times in zip(self.sample.sequences, self.sample.times):
			for i in range(0,len(seq),2):
				if not seq[:i+1] in states_lbl:
					states_lbl.append(seq[:i+1])
					states.append( IOFPTA_state(seq[i], times, seq[:i+1]))
				else:
					states[states_lbl.index(seq[:i+1])].counterAdd(times)
		#sorting states
		states_lbl.sort()
		for s in states:
			s.setId(states_lbl.index(s.label))
		states_sorted = [None]*len(states)	
		for s in states:
			states_sorted[s.id] = s
		#init states_transitions

		initial = []
		for s1 in range(len(states_sorted)):
			len_s1 = len(states_sorted[s1].label)
			if len_s1 == 1:
				initial.append(s1)
				
			s2 = s1 + 1
			while s2 < len(states_sorted):
				if len(states_sorted[s2].label) < len_s1 + 2: # too short
					s2 += 1
				elif len(states_sorted[s2].label) > len_s1 + 2: # too long
					s2 += 1
				elif states_sorted[s2].label[:-2] != states_sorted[s1].label: # not same prefix
					s2 += 1
				else: # OK
					act = states_sorted[s2].label[-2]
					states_sorted[s1].transitionsAdd(act, states_sorted[s2].counter, s2)
					s2 += 1

		res =  IOFPTA(states_sorted,self.observations,self.actions)
		for i in initial:
			res.addInitialState(i)
		return res

	def fit(self,sample:Set,epsilon:float,
	 		stormpy_output: bool = True, output_file_prism : str = None):
		"""
		Fits the model according to ``traces``.

		Parameters
		----------
		sample : Set
			A trainig set.
		epsilon : float
			Espilon parameter for the compatibility test.
			Should be between 0 and 1 (0 excluded).
		stormpy_output: bool, optional
			If set to True the output model will be a Stormpy sparse model.
			Default is True.
		output_file_prism : str, optional
			If set, the output model will be saved in a prism file at this
			location. Otherwise the output model will not be saved.
			This parameter is ignored if the model under learning is a HMM
			or a GoHMM.

		Returns
		-------
		MDP or stormpy.SparseMdp
			The fitted MDP.
			If `stormpy_output` is set to `False` or if stormpy is not available on
			the machine it returns a `jajapy.MDP`, otherwise it returns a `stormpy.SparseMdp`
		"""
		self.actions, self.observations = sample.getActionsObservations()
		self._initialize(sample,epsilon,self.actions,self.observations)
		red = [0]
		blue = []
		for i in self.a.initial:
			blue += self.a.states[i].successors()

		while len(blue) != 0:
			state_b = blue[0]
			merged = False
			
			for state_r in red:
				if self.t.compatible(state_r,state_b,epsilon):
					self.a.merge(state_r,state_b)
					merged = True
					break

			if not merged:
				red.append(state_b)
			blue.remove(state_b)

			for state_r in red:
				for state_b in self.a.states[state_r].successors():
					if state_b not in red:
						blue.append(state_b)

			blue = list(set(blue))
			blue.sort()

		m = self.a.cleanMDP(red)

		try:
			from ..with_stormpy import jajapyModeltoStormpy
			stormpy_installed = True
		except ModuleNotFoundError:
			stormpy_installed = False
		if stormpy_output and not stormpy_installed:
			print("WARNING: stormpy not found. The output model will not be a stormpy sparse model")
			stormpy_output = False


		if output_file_prism:
			self.h.savePrism(output_file_prism)
		
		if stormpy_output:
			return jajapyModeltoStormpy(m)
		else:
			return m


import unittest
from ..base.tools import *

class ToolsTestclass(unittest.TestCase):

	def test_resolveRandom(var):
		probabilities = [0.3,0.6,0.1]
		nb_trials = 100000
		results = [0.0,0.0,0.0]
		for i in range(nb_trials):
			results[resolveRandom(probabilities)] += 1/nb_trials
		for i in range(len(probabilities)):
			var.assertAlmostEqual(probabilities[i],results[i],delta=3)
	
	def test_randomProbabilities(var):
		var.assertRaises(ValueError, randomProbabilities, 0)
		var.assertRaises(TypeError, randomProbabilities, 0.2)
		for length in range(1, 6):
			p = randomProbabilities(length)
			var.assertEqual(len(p),length)
			var.assertAlmostEqual(sum(p),1.0)
			for i in range(length):
				var.assertGreaterEqual(p[i],0.0)


if __name__ == "__main__":
	unittest.main()


from .Scheduler import UniformScheduler, MemorylessScheduler
from ..base.BW import BW, NB_PROCESS
from ..base.Set import Set
from ..base.tools import resolveRandom
from multiprocessing import Pool
from random import random
from numpy import zeros, dot, array, argmin, inf
from sys import platform
from datetime import datetime

class ActiveLearningScheduler:
	"""
	Class for a scheduler used during the active learning sampling.
	"""
	def __init__(self,memoryless_scheduler: MemorylessScheduler,m) -> None:
		"""
		Create an ActiveLearningScheduler.

		Parameters
		----------
		memoryless_scheduler : MemorylessScheduler
			A memoryless scheduler generated by Active_BW_MDP._strategy
		m : MDP
			The current hypothesis.
		"""
		self.m = m
		self.nb_states = self.m.nb_states
		self.memoryless_scheduler = memoryless_scheduler
		self.reset()
		
	def reset(self) -> None:
		"""
		Resets the scheduler. Should be done before genereting a new trace.
		"""
		self.last_obs = None
		self.last_act = None
		self.alpha_matrix = array([self.m.pi(s) for s in range(self.nb_states)])

	def getAction(self) -> str:
		"""
		Returns the action chosen by the scheduler

		Returns
		-------
		str
			Action chosen by the scheduler.
		"""
		if self.last_obs != None:
			alpha_matrix = zeros(self.nb_states)
			for s in range(self.nb_states):
				p = array([self.m.tau(ss,self.last_act,s,self.last_obs) for ss in range(self.nb_states)])
				alpha_matrix[s] = dot(self.alpha_matrix,p)
			self.alpha_matrix = alpha_matrix

		tot = self.alpha_matrix.sum()
		if tot <= 0.0:
			t = [1/len(t) for i in self.alpha_matrix]
		else:
			t = [i/tot for i in self.alpha_matrix]
		s_i = resolveRandom(t)
		act = self.memoryless_scheduler.getAction(s_i)
		self.last_act = act
		return act

	def addObservation(self,obs: str) -> None:
		"""
		Updates the scheduler according to the given observation ``obs``.

		Parameters
		----------
		obs : str
			An observation.
		"""
		self.last_obs = obs

class Active_BW_MDP(BW):
	"""
	Class for the Active Baum-Welch algorithm on MDP.
	This algorithm is described here:
	https://arxiv.org/pdf/2110.03014.pdf
	"""
	def __init__(self):
		super().__init__()

	def fit(self,traces: Set, sul, lr, nb_iterations: int, nb_sequences: int,
			sequence_length: int = None, epsilon_greedy: float = 0.9,
			initial_model=None, nb_states: int=None,
			random_initial_state: bool=True, output_file: str=None,
			epsilon: float=0.01, max_it: int=inf,pp: str='',
			verbose: bool = True, return_data: bool= False, stormpy_output: bool = True):
		"""
		Fits the model according to ``traces``.

		Parameters
		----------
		traces : Set
			Training set.
		sul : MDP
			The system under learning.
			Used to generate new sequences.
		lr : float, or ``0``, or ``"dynamic"``
			Learning rate. If ``lr=0`` the current hypothesis will be updated
			on ``traces`` plus all the active samples. If `lr` is a float, the
			current hypothesis will be merged with an updated version of it.
			The updated version is a version of the current hypothesis fitted
			on the last active sample. If `lr="dynamic"` the learning rate
			decrease at each iteration.
		nb_iterations: int
			Number of active learning iterations.
		nb_sequences: int
			Size of one active sample. I.e: number of sequences generated at
			each active learning iteration.
		sequence_length: int, optional.
			Size of one sequence generated during an active learning iteration.
			If not set it will be equal to the size of the first sequence in
			``traces``.
		epsilon_greedy: float, optional.
			It will use the active learning scheduler with probability
			``epsilon_greedy`` while genereting the active learning sample, and
			with probability `1-epsilon_greedy` it will use an uniform scheduler.
			By default 0.9
		initial_model : MDP, optional.
			First hypothesis. If not set it will create a random MDP with
			``nb_states`` states. Should be set if ``nb_states`` is not set.
		nb_states: int
			If ``initial_model`` is not set it will create a random MDP with
			``nb_states`` states. Should be set if ``initial_model`` is not set.
			Default is None.
		random_initial_state: bool, optional.
			If ``initial_model`` is not set it will create a random MDP with
			random initial state according to this sequence of probabilities.
			Should be set if ``initial_model`` is not set.
			Default is False.
		output_file : str, optional
			If set path file of the output model. Otherwise the output model
			will not be saved into a text file.
		epsilon : float, optional
			The learning process stops when the difference between the
			loglikelihood of the training set under the two last hypothesis is
			lower than ``epsilon``. The lower this value the better the output,
			but the longer the running time. By default 0.01.
		max_it: int
			Maximal number of iterations for the passive learning round.
			The algorithm will stop after `max_it` iterations.
			Default is infinity.
		pp : str, optional
			Will be printed at each iteration. By default ''
		verbose: bool, optional
			Print or not a small recap at the end of the learning.
			Default is True.
		return_data: bool, optional
			If set to True, a dictionary containing following values will be
			returned alongside the hypothesis once the learning is done.
			'learning_rounds', 'learning_time', 'training_set_loglikelihood',
			Default is False.
		stormpy_output: bool, optional
			If set to True the output model will be a Stormpy sparse model.
			Default is True.

		Returns
		-------
		MDP
			fitted MDP.
		"""
		if stormpy_output:
			try:
				from ..with_stormpy import jajapyModeltoStormpy
			except ModuleNotFoundError:
				print("WARNING: stormpy not found. The output model will not be a stormpy sparse model")
				stormpy_output = False

		counter = 0
		self.sul = sul
		start_time = datetime.now()
		_, info = super().fit(traces, initial_model,nb_states,
					random_initial_state, output_file,epsilon,
					max_it,"Passive iteration:"+pp,
					return_data=True, stormpy_output=False)
		counter += info['learning_rounds']
	
		total_traces = traces
		
		if sequence_length == None:
			sequence_length = int(len(traces.sequences[0])/2)

		c = 1
		while c <= nb_iterations :
			traces = self._addTraces(sequence_length,nb_sequences,total_traces,epsilon_greedy)
			total_traces.addSet(traces)
			
			if lr == 0:
				_, info = super().fit(total_traces, initial_model=self.h,
							output_file=output_file, epsilon=epsilon,
							pp="Active iteration "+str(c)+"/"+str(nb_iterations)+": "+pp,
							return_data=True, stormpy_output=False, verbose=False)
				counter += info['learning_rounds']
			else:
				if lr == "dynamic":
					lr_it = sum(traces.times)/(sum(total_traces.times)-sum(traces.times))
				else:
					lr_it = lr
				old_h = self.h
				_, info = super().fit(traces, initial_model=self.h,
							output_file=output_file, epsilon=epsilon,
							pp="Active iteration "+str(c)+"/"+str(nb_iterations)+": "+pp,
							return_data=True, stormpy_output=False, verbose=False)
				counter += info['learning_rounds']
				self._mergeModels(old_h,lr_it)

			c += 1
		
		running_time = datetime.now()-start_time
		running_time = running_time.total_seconds()

		if verbose:
			self._endPrint(counter, running_time)

		if stormpy_output:
			self.h = jajapyModeltoStormpy(self.h)
			
		if return_data:
			info = {"learning_rounds":counter,
					"learning_time":running_time,
					"training_set_loglikelihood":self.h.logLikelihood(total_traces)}
			return self.h, info

		return self.h

	def _mergeModels(self,old_h,lr: float) -> None:
		"""
		Merges the model ``self.h`` with model ``old_h`` using the learning
		rate ``lr``.
		WARNING: we assume that self.h.alphabet == old_h.alphabet and
		self.h.actions == old_h.actions.

		Parameters
		----------
		old_h : MDP
			A MDP to merge with ``self.h``. 
		lr : float
			Learning rate. It is the weight of ``self.h`` in the output model.
		"""
		self.h.matrix = lr*self.h.matrix + (1-lr)*old_h

	def _addTraces(self,sequence_length: int,nb_sequence: int,
				  traces: Set,epsilon_greedy: float) -> list:
		"""
		Generates an active learning sample.

		Parameters
		----------
		sequence_length : int
			Length of each sequence in the sample.
		nb_sequence : int
			Number of sequences in the sample.
		traces : Set
			Training set until now. Used to compute the active learning
			sampling strategy.
		epsilon_greedy : It will use the active learning scheduler with probability
			``epsilon_greedy`` and use an uniform scheduler with probability
			`1-epsilon_greedy`.
			
		Returns
		-------
		list
			The active learning sample
		"""
		memoryless_scheduler = self._strategy(traces)
		scheduler_exploit = ActiveLearningScheduler(memoryless_scheduler,self.h)
		scheduler_explore = UniformScheduler(self.h.getActions())

		traces = Set([],t=1)
		for n in range(nb_sequence):			
			if random() > epsilon_greedy:
				seq = self.sul.run(sequence_length,scheduler_explore)
			else:
				seq = self.sul.run(sequence_length,scheduler_exploit)

			if not seq in traces.sequences:
				traces.sequences.append(seq)
				traces.times.append(0)
			traces.times[traces.sequences.index(seq)] += 1
		return traces

	def _strategy(self,traces:Set) -> MemorylessScheduler:
		if platform != "win32":
			p = Pool(processes = NB_PROCESS)
			tasks = []
			for seq,times in zip(traces.sequences,traces.times):
				tasks.append(p.apply_async(self._computeProbas, [seq, times,]))
			temp = array([res.get() for res in tasks])
		else:
			temp = array([self._computeProbas(seq, times) for seq,times in zip(traces.sequences,traces.times)])
		temp = temp.sum(axis=0)
		scheduler = [self.h.getActions()[argmin(temp[s])] for s in range(self.nb_states)]
		return MemorylessScheduler(scheduler)

	def _computeProbas(self,seq:list,time:int) -> array:
		sequence_actions = [seq[i+1] for i in range(0,len(seq)-1,2)]
		sequence_obs = [seq[i] for i in range(0,len(seq)-1,2)] + [seq[-1]]
		alpha_matrix = self._computeAlphas(sequence_obs,sequence_actions).T[:-1].T
		res = zeros(shape=(len(self.actions),self.nb_states))
		fact = alpha_matrix.sum(axis=0) # fact[t] proba to generate sequence_obs[:t]

		for ia,a in enumerate(self.actions):
			arr_dirak = array([1.0 if t == a else 0.0 for t in sequence_actions])
			arr_dirak = arr_dirak*time/fact
			res[ia] = dot(alpha_matrix,arr_dirak)
		return res.T



from ..base.tools import resolveRandom, randomProbabilities, checkProbabilities
from ..base.Base_MC import *
from ast import literal_eval
from numpy import ndarray, array, zeros, vstack, hstack, newaxis, append, where

class MC(Base_MC):
	def __init__(self, matrix: ndarray, labelling: list, name: str ="unknown_MC") -> None:
		"""
		Creates an MC.

		Parameters
		----------
		matrix : ndarray
			A (N x N) ndarray (with N the nb of states).
			Represents the transition matrix.
			`matrix[s1][s2]` is the probability of moving from `s1` to `s2`.
		labelling: list of str
			A list of N observations (with N the nb of states).
			If `labelling[s] == o` then state of ID `s` is labelled by `o`.
			Each state has exactly one label.
		name : str, optional
			Name of the model.
			Default is "unknow_MC"
		"""	
		super().__init__(matrix,labelling,name)
		self.model_type = MC_ID
		for i in range(self.nb_states):
			if not checkProbabilities(matrix[i]):
				msg = "The probability to take a transition from state "
				msg+= str(i)+" should be 1.0, here it's "+str(matrix[i].sum())
				raise ValueError(msg)
			
	def tau(self,s1: int, s2: int, obs: str) -> float:
		"""
		Returns the probability of moving from state `s1` to `s2` seeing label `obs`.
		(i.e. if `s1` is not labelled with `obs` the probability is 0.0).

		Parameters
		----------
		s1: int
			source state ID.
		s2: int
			destination state ID.
		obs: str
			seen label.
		
		Returns
		-------
		float
			probability of moving from state `s1` to `s2` seeing label `obs`.
		
		Example
		-------
		>>> model.tau(0,1,'a')
		0.6
		>>> model.getLabel(0)
		'a'
		>>> model.tau(0,1,'b')
		0.0
		>>> model.getLabel(1)
		'b'
		"""
		self._checkStateIndex(s1)
		self._checkStateIndex(s2)
		if obs != self.labelling[s1]:
			return 0.0
		return self.matrix[s1][s2]
	
	def a(self,s1: int,s2: int) -> float:
		"""
		Returns the probability of moving from state `s1` to state `s2`.

		Parameters
		----------
		s1 : int
			ID of the source state.		
		s2 : int
			ID of the destination state.
		
		Returns
		-------
		float
			Probability of moving from state `s1` to state `s2`.
		
		Example
		-------
		>>> model.a(0,1)
		0.6
		"""
		self._checkStateIndex(s1)
		self._checkStateIndex(s2)
		return self.matrix[s1][s2]

	def next(self,state: int) -> tuple:
		"""
		Return a state-observation pair according to the distributions described by `matrix`

		Returns
		-------
		output : (int, str)
			A state-observation pair.
			
		Example
		-------
		>>> model.next(0)
		(1,'a')
		>>> model.getLabel(0)
		'a'
		>>> model.next(0)
		(1,'a')
		>>> model.next(0)
		(2,'a')
		>>> model.a(0,1)
		0.6
		>>> model.a(0,2)
		0.4
		"""
		c = resolveRandom(self.matrix[state])
		return (c, self.labelling[state])

	def run(self, number_steps: int, current: int = -1) -> list:
		"""
		Simulates a run of length ``number_steps`` of the model and return the
		sequence of observations generated.
		
		Parameters
		----------
		number_steps: int
			length of the simulation.
		current : int, optional.
			If current it set, it starts from the state `current`.
			Otherwise it starts from an initial state.

		Returns
		-------
		output: list of str
			trace generated by the run.
		"""
		output = []
		if current == -1:
			current = resolveRandom(self.initial_state)

		while len(output) < number_steps:
			[next_state, symbol] = self.next(current)
			output.append(symbol)
			current = next_state
		
		output.append(self.labelling[current])
		return output

	def save(self, file_path:str) -> None:
		"""Save the model into a text file.

		Parameters
		----------
		file_path : str
			path of the output file.
		
		Examples
		--------
		>>> model.save("my_model.txt")
		"""
		f = open(file_path, 'w')
		f.write("MC\n")
		super()._save(f)

	def _stateToString(self,state:int) -> str:
		res = "----STATE "+str(state)+"--"+self.labelling[state]+"----\n"
		for j in range(len(self.matrix[state])):
			if self.matrix[state][j] > 0.0001:
				res += "s"+str(state)+" -> s"+str(j)+" : "+str(round(self.matrix[state][j],5))+'\n'
		return res
	
	def savePrism(self,file_path:str) -> None:
		"""
		Save this model into `file_path` in the Prism format.

		Parameters
		----------
		file_path : str
			Path of the output file.
		"""
		f = open(file_path,'w')
		f.write("dtmc\n\n")
		super()._savePrism(f)
	
	def _savePrismMatrix(self,f) -> None:
		for s1 in range(self.nb_states):
			if (self.matrix[s1]!=0).any():
				res = '\t[] s='+str(s1)+' ->'
				for s2 in where(self.matrix[s1] != 0.0)[0]:
					res += ' '+str(self.matrix[s1,s2])+":(s'="+str(s2)+") +"
				res = res[:-2]+';\n'
				f.write(res)
				

def loadMC(file_path: str) -> MC:
	"""
	Load an MC saved into a text file.

	Parameters
	----------
	file_path : str
		Location of the text file.
	
	Returns
	-------
	output : MC
		The MC saved in `file_path`.
	
	Examples
	--------
	>>> model = loadMC("my_model.txt")
	"""
	f = open(file_path,'r')
	l = f.readline()[:-1] 
	if l != "MC":
		msg = "ERROR: this file doesn't describe an MC: it describes a "+l
		raise ValueError(msg)
	labelling = literal_eval(f.readline()[:-1])
	name = f.readline()[:-1]
	initial_state = array(literal_eval(f.readline()[:-1]))
	matrix = literal_eval(f.readline()[:-1])
	matrix = array(matrix)
	f.close()
	return MC(matrix, labelling, name)

def MC_random(nb_states: int, labelling: list, random_initial_state: bool=True) -> MC:
	"""
	Generate a random MC.

	Parameters
	----------
	number_states : int
		Number of states.
	labelling : list of str
		List of observations.
	random_initial_state: bool, optional
		If set to True we will start in each state with a random probability,
		otherwise we will always start in state 0.
		Default is True.
	
	Returns
	-------
	MC
		A pseudo-randomly generated MC.

	Examples
	--------
	>>> model = MC_random(2,['a','b'],False)
	>>> print(model)
	Name: MC_random_2_states
	Initial state: s2
	----STATE 0--a----
	s0 -> s0 : 0.625
	s0 -> s1 : 0.375
	----STATE 1--b----
	s1 -> s0 : 0.9
	s1 -> s1 : 0.1
	----STATE 2--init----
	s2 -> s0 : 1.0
	"""
	matrix = []
	for _ in range(nb_states):
		matrix.append(append(randomProbabilities(nb_states),0.0))

	if random_initial_state:
		matrix.append(append(randomProbabilities(nb_states),0.0))
	else:
		matrix.append(array([1.0]+[0.0 for i in range(nb_states)]))
	matrix = array(matrix)
	labelling = labelsForRandomModel(nb_states,labelling)
	return MC(matrix, labelling,"MC_random_"+str(nb_states)+"_states")

def createMC(transitions: list, labelling: list, initial_state, name: str ="unknown_MC") -> MC:
	"""
	An user-friendly way to create a MC.

	Parameters
	----------
	transitions : [ list of tuples (int, int, float)]
		Each tuple represents a transition as follow: 
		(source state ID, destination state ID, probability).
	labelling: list of str
		A list of N observations (with N the nb of states).
		If `labelling[s] == o` then state of ID `s` is labelled by `o`.
		Each state has exactly one label.
	initial_state : int or list of float
		Determine which state is the initial one (then it's the id of the
		state), or what are the probability to start in each state (then it's
		a list of probabilities).
	name : str, optional
		Name of the model.
		Default is "unknow_MC"
	
	Returns
	-------
	MC
		the MC describes by `transitions`, `labelling`, and `initial_state`.
	
	Examples
	--------
	>>> model = createMC([(0,1,1.0),(1,0,0.6),(1,1,0.4)],['b','a'],0,"My_MC")
	>>> print(model)
	Name: My_MC
	Initial state: s2
	----STATE 0--b----
	s0 -> s1 : 1.0
	----STATE 1--a----
	s1 -> s0 : 0.6
	s1 -> s1 : 0.4
	----STATE 2--init----
	s2 -> s0 : 1.0
	"""
	if 'init' in labelling:
		msg =  "The label 'init' cannot be used: it is reserved for initial states."
		raise SyntaxError(msg)
	
	states = list(set([i[0] for i in transitions]+[i[1] for i in transitions]))
	states.sort()
	nb_states = len(states)
	
	if nb_states > len(labelling):
		raise ValueError("All states are not labelled (the labelling list is too small).")
	elif nb_states < len(labelling):
		print("WARNING: the labelling list is bigger than the number of states")

	res = zeros((nb_states,nb_states))
	for t in transitions:
		res[states.index(t[0])][states.index(t[1])] = t[2]
	
	labelling.append('init')
	res = vstack((res,zeros(len(res))))
	res = hstack((res,zeros(len(res))[:,newaxis]))
	if type(initial_state) == int:
		res[-1][initial_state] = 1.0
	else:
		if type(initial_state) == ndarray:
			initial_state = initial_state.tolist()
		if len(initial_state) != nb_states:
			raise ValueError('The size of the initial_state list must be equal to the number of states')
		res[-1] = array(initial_state+[0.0])

	return MC(res, labelling, name)



import stormpy as st
from numpy import zeros, array, newaxis, reshape, vstack, concatenate, hstack, newaxis, nan, full
from ..mc.MC import MC
from ..mdp.MDP import MDP
from ..ctmc.CTMC import CTMC
from ..pctmc.PCTMC import PCTMC
from copy import deepcopy
import os
from sympy import symbols, sympify

def stormpyModeltoJajapy(h,actions_name:list = [],from_prism=False):
	"""
	Given a stormpy.SparseCtmc, stormpy.SparseDtmc, stormpy.SparseMdp, or
	stormpy.SparseParametricCtmc, it returns the equivalent jajapy model.
	The output object will be a jajapy.MC, jajapy.CTMC, jajapy.MDP or
	jajapy.PCTMC depending on the input.

	Parameters
	----------
	h : stormpy.SparseCtmc, stormpy.SparseDtmc, stormpy.SparseMdp or stormpy.SparseParametricCtmc
		The model to convert.
	
	actions_name : list of str, optional.
		If the model is an MDP, the name of the actions in the output
		model will be the one in this list. Otherwise they will be
		`a0,a1,a2,...`.

	Returns
	-------
	jajapy.MC, jajapy.CTMC, jajapy.MDP or jajapy.PCTMC
		The same model in jajapy format.
	"""
	if type(h) == st.SparseDtmc:
		ty = 0
	elif type(h) == st.SparseCtmc:
		ty = 1
	elif type(h) == st.SparseMdp:
		ty = 2
	#elif type(h) == st.SparseParametricDtmc:
	#	ty = 3
	elif type(h) == st.SparseParametricCtmc:
		ty = 4
	else:
		raise TypeError(str(type(h))+' cannot be translated to Jajapy model.')
	
	labelling = [None for _ in range(len(h.states))]
	if ty == 2:
		actions = []
		for s in h.states:
			for a in s.actions:
				if len(actions_name) <= int(str(a)):
					actions.append('a'+str(a))
				else:
					actions.append(actions_name[int(str(a))])
		actions = list(set(actions))
		matrix = zeros((len(h.states),len(actions),len(h.states)))
	elif ty == 0 or ty == 1:
		matrix = zeros((len(h.states),len(h.states)))
	elif ty == 3 or ty == 4:
		matrix = zeros((len(h.states),len(h.states)),dtype='uint16')
		p_str = []
		p_v = {}
		p_i = []
		t_expr = [sympify(0.0)]

	add_init_state = None
	for si,s in enumerate(h.states):
		c = si
		temp = list(s.labels)
		if "deadlock" in temp:
			temp.remove("deadlock")
		temp.sort()
		if len(temp) == 0:
			labelling[si] = "empty"
		elif 'init' in temp and len(temp) > 1:
			temp.remove("init")
			labelling.append("init")
			labelling[si] = '_'.join(list(temp))
			add_init_state = c
		else:
			labelling[si] = '_'.join(list(temp))

		for a in s.actions:
			for t in a.transitions:
				dest = t.column
				t_val = t.value()
				if ty == 2:
					matrix[c][int(str(a))][dest] = t_val
				elif ty == 1 or ty == 0:
					matrix[c][dest] = t_val
				else:
					ps = [i.name for i in list(t_val.gather_variables())]
					if len(ps) == 1:
						ps = [symbols(ps[0])]
					elif len(ps) > 1:
						ps = list(symbols(" ".join(ps)))
					for v in ps:
						v = v.name
						if not v in p_str:
							p_str.append(v)
							p_i.append([])
							p_v[v] = nan
						p_i[p_str.index(v)].append([c,dest])
					t_val = sympify(str(t_val))
					if t_val.is_real or not t_val in t_expr:
						matrix[c][dest] = len(t_expr)
						t_expr.append(t_val)
					else:
						matrix[c][dest] = t_expr.index(t_val)
		if ty == 1 and not from_prism:
			matrix[c] *= h.exit_rates[si]
	
	if add_init_state != None:
		#matrix = vstack((matrix,matrix[add_init_state]))
		if ty == 2:
			matrix = vstack((matrix,zeros((1,matrix.shape[1],matrix.shape[2]))))
			matrix[-1].T[add_init_state] = full(matrix.shape[1],1.0)
			matrix = concatenate((matrix,zeros((matrix.shape[0],matrix.shape[1],1))),axis=2)
		elif ty == 1 or ty == 0:
			matrix = vstack((matrix,zeros((matrix.shape[0]))))
			matrix[-1][add_init_state] = 1.0
			matrix = hstack((matrix,zeros(len(matrix))[:,newaxis]))
		else:
			matrix = vstack((matrix,zeros((matrix.shape[0]),dtype='uint16')))
			t_val = sympify('1.0')
			matrix[-1][add_init_state] = len(t_expr)
			t_expr.append(t_val)
			matrix = hstack((matrix,zeros(len(matrix),dtype=('uint16'))[:,newaxis]))

	if ty == 0:
		return MC(matrix, labelling)
	elif ty == 1:
		return CTMC(matrix, labelling)
	elif ty == 2:
		return MDP(matrix,labelling,actions)
	#elif ty == 3:
	#	return PMC(matrix,labelling,p_v,p_i,p_str)
	elif ty == 4:
		return PCTMC(matrix,labelling,t_expr,p_v,p_i,p_str)

def jajapyModeltoStormpy(h):
	"""
	Given a jajapy.MC, a jajapy.CTMC, a jajapy.MDP or an instantiated
	jajapy.PCTMC, it returns the equivalent stormpy sparse model.
	The output object will be a stormpy.SparseCtmc, stormpy.SparseDtmc,
	stormpy.SparseMdp, or stormpy.SparseParametricCtmc depending on the input.

	Parameters
	----------
	h : jajapy.MC, jajapy.CTMC, jajapy.MDP or instantiated jajapy.PCTMC
		The model to convert.

	Returns
	-------
	stormpy.SparseCtmc, stormpy.SparseDtmc, stormpy.SparseMdp or stormpy.SparseParametricCtmc
		The same model in stormpy format.
	"""
	if type(h) == MDP:
		return MDPtoStormpy(h)
	elif type(h) == CTMC:
		return CTMCtoStormpy(h)
	elif type(h) == MC:
		return MCtoStormpy(h)
	elif type(h) == PCTMC:
		try:
			h = PCTMCtoCTMC(h)
		except ValueError:
			raise ValueError("Cannot convert non-instantiated PCTMC to Stormpy.")
		return CTMCtoStormpy(h)
	else:
		raise TypeError(str(type(h))+' cannot be translated to a stormpy sparse model.')

def _buildStateLabeling(h):
	state_labelling = st.storage.StateLabeling(h.nb_states)
	for o in h.getAlphabet():
		state_labelling.add_label(o)
	for s in range(h.nb_states):
		state_labelling.add_label_to_state(h.labelling[s],s)
	return state_labelling

def MDPtoStormpy(h):
	"""
	Given a jajapy.MDP, it returns the equivalent stormpy sparse model.
	The output object will be a stormpy.SparseMdp.

	Parameters
	----------
	h : jajapy.MDP
		The model to convert.

	Returns
	-------
	stormpy.SparseMdp
		The same model in stormpy format.
	"""
	state_labelling = _buildStateLabeling(h)
	nb_actions = len(h.getActions())
	transition_matrix = h.matrix
	transition_matrix = reshape(transition_matrix.flatten(),(h.nb_states*nb_actions,h.nb_states))
	transition_matrix =  st.build_sparse_matrix(transition_matrix,[nb_actions*i for i in range(h.nb_states)])
	choice_labelling = st.storage.ChoiceLabeling(h.nb_states*nb_actions)
	for ia,a in enumerate(h.getActions()):
		choice_labelling.add_label(a)
		choice_labelling.add_label_to_choice(a,ia)
	reward_models = {}
	action_reward = [-1.0 for _ in range(len(transition_matrix))]
	reward_models["nb_executed_actions"] = st.SparseRewardModel(optional_state_action_reward_vector = action_reward)
	components = st.SparseModelComponents(transition_matrix=transition_matrix,
										  state_labeling=state_labelling,
										  reward_models=reward_models)
	components.choice_labeling = choice_labelling
	mdp = st.storage.SparseMdp(components)
	return mdp

def MCtoStormpy(h):
	"""
	Given a jajapy.MC, it returns the equivalent stormpy sparse model.
	The output object will be a stormpy.SparseDtmc.

	Parameters
	----------
	h : jajapy.MC
		The model to convert.

	Returns
	-------
	stormpy.SparseDtmc
		The same model in stormpy format.
	"""
	state_labelling = _buildStateLabeling(h)
	transition_matrix = h.matrix
	transition_matrix =  st.build_sparse_matrix(transition_matrix)
	components = st.SparseModelComponents(transition_matrix=transition_matrix,
										  state_labeling=state_labelling)
	mc = st.storage.SparseDtmc(components)
	return mc

def CTMCtoStormpy(h):
	"""
	Given a jajapy.CTMC, it returns the equivalent stormpy sparse model.
	The output object will be a stormpy.SparseCtmc.

	Parameters
	----------
	h : jajapy.CTMC
		The model to convert.

	Returns
	-------
	stormpy.SparseCtmc
		The same model in stormpy format.
	"""
	state_labelling = _buildStateLabeling(h)
	transition_matrix = deepcopy(h.matrix)
	e = array([h.e(s) for s in range(h.nb_states)])
	transition_matrix /= e[:,newaxis]
	transition_matrix =  st.build_sparse_matrix(transition_matrix)
	components = st.SparseModelComponents(transition_matrix=transition_matrix,
										  state_labeling=state_labelling,
										  rate_transitions=True)
	components.exit_rates = e
	ctmc = st.storage.SparseCtmc(components)
	return ctmc

def PCTMCtoCTMC(h: PCTMC) -> CTMC:
	"""
	Translates a given instantiated PCTMC to an equivalent CTMC.

	Parameters
	----------
	h : PCTMC
		An instantiated PCTMC.

	Returns
	-------
	CTMC
		The equivalent CTMC.

	Raises
	------
	ValueError
		If `h` is a non-instantiated PCTMC.
	"""
	if not h.isInstantiated():
		raise ValueError("Cannot convert non-instantiated PCTMC to CTMC.")
	res = zeros(h.matrix.shape)
	for s in range(h.nb_states):
		for ss in range(h.nb_states):
			res[s,ss] = h.transitionValue(s,ss)
	return CTMC(res, h.labelling, h.name)

def loadPrism(path: str):
	"""
	Load the model described in file `path` under Prism format.
	Remark: this function uses the stormpy parser for Prism file.

	Remarks
	-------
	For technical reason, this function clear the terminal on usage.

	Parameters
	----------
	path : str
		Path to the Prism model to load.

	Returns
	-------
	jajapy.MC, jajapy.CTMC, jajapy.MDP or jajapy.PCTMC
		A jajapy model equivalent to the model described in `path`.
	"""
	try:
		prism_program = st.parse_prism_program(path,False)
	except RuntimeError:
		prism_program = st.parse_prism_program(path,True)
	try:
		stormpy_model = st.build_model(prism_program)
	except RuntimeError:
		stormpy_model = st.build_parametric_model(prism_program)

	if os.name != "nt":
		os.system('clear')
	else:
		os.system('cls')

	jajapy_model = stormpyModeltoJajapy(stormpy_model,from_prism=True)
	return jajapy_model


from .tools import resolveRandom, checkProbabilities
from .Model import Model, HMM_ID, GOHMM_ID

class Base_HMM(Model):
	"""
	Abstract class that represents a model.
	Is inherited by HMM and GoHMM.
	Should not be instantiated itself!
	"""

	def __init__(self,matrix, initial_state, name):
		"""
		Creates an abstract model for HMM and GoHMM.

		Parameters
		----------
		matrix : ndarray
			Represents the transition matrix.
			`matrix[s1][s2]` is the probability of moving from `s1` to `s2`.
		initial_state : int or list of float
			Determine which state is the initial one (then it's the id of the
			state), or what are the probability to start in each state (then it's
			a list of probabilities).
		name : str, optional
			Name of the model.
		"""
		super().__init__(matrix,initial_state,name)
		for i in range(self.nb_states):
			if not checkProbabilities(matrix[i]):
				raise ValueError("The probability to take a transition from state",i,"should be 1.0, here it's",matrix[i].sum())
				

	def a(self,s1: int,s2: int) -> float:
		"""
		Returns the probability of moving from state `s1` to state `s2`.
		If `s1` or `s2` is not a valid state ID it returns 0.

		Parameters
		----------
		s1 : int
			ID of the source state.		
		s2 : int
			ID of the destination state.
		
		Returns
		-------
		float
			Probability of moving from state `s1` to state `s2`.
		
		Examples
		--------
		>>> model.a(0,1)
		0.5
		>>> model.a(0,0)
		0.0
		"""
		self._checkStateIndex(s1)
		self._checkStateIndex(s2)
		return self.matrix[s1][s2]

	def tau(self, s1: int, s2: int, obs) -> float:
		"""
		Return the probability of generating from state `s1` observation `obs` and moving to state `s2`.
		If `s1` or `s2` is not a valid state ID it returns 0.

		Parameters
		----------
		s1 : int
			ID of the source state.
		s2 : int
			ID of the destination state.
		obs : str or list
			An observation. `str` if the model is an HMM,`list` if it's a GoHMM

		Returns
		-------
		float
			The probability of generating from state `s1` observation `obs` and moving to state `s`.
		"""
		return self.a(s1,s2)*self.b(s1,obs)

	def _stateToString(self, state: int) -> str:
		res = "----STATE s"+str(state)+"----\n"
		for j in range(len(self.matrix[state])):
			if self.matrix[state][j] > 0.0001:
				res += "s"+str(state)+" -> s"+str(j)+" : "+str(self.matrix[state][j])+'\n'
		res += "************\n"
		return res
	
	def _save(self,f) -> None:
		f.write(str(self.output.tolist()))
		f.write('\n')
		super()._save(f)

	def next_obs(self, state: int) -> str:
		"""
		Generates one observation according to the distribution described by `self.output_matrix`.
		
		Returns
		-------
		str
			An observation.
		"""
		c = resolveRandom(self.output[state].flatten())
		return self.alphabet[c]

	def next_state(self, state: int) -> int:
		"""
		Returns one state ID at random according to the distribution described 
		by the `self.matrix`.
		
		Parameters
		----------
		state : int
			ID of the source state.	
		
		Returns
		-------
		int
			A state ID.
		"""
		c = resolveRandom(self.matrix[state].flatten())
		return c

	def next(self, state: int) -> list:
		"""
		Returns a state-observation pair according to the distributions
		described by `self.matrix[s]` and `self.output[s]`.

		Parameters
		----------
		state : int
			ID of the source state.

		Returns
		-------
		output : [int, list of floats]
			A state-observation pair.
		"""
		return [self.next_state(state),self.next_obs(state)]

